from scNodes.core.node import *
from scNodes.core.reconstruction import *


def create():
    return ReconstructionRendererNode()


class ReconstructionRendererNode(Node):
    title = "Render reconstruction"
    group = "PSF-fitting reconstruction"
    colour = (243 / 255, 0 / 255, 80 / 255, 1.0)
    size = 250
    COLOUR_MODE = ["RGB, LUT"]
    sortid = 1005

    DISABLE_FRAME_INFO_WINDOW = True

    def __init__(self):
        super().__init__()

        self.reconstruction_in = ConnectableAttribute(ConnectableAttribute.TYPE_RECONSTRUCTION, ConnectableAttribute.INPUT, parent = self)
        self.image_out = ConnectableAttribute(ConnectableAttribute.TYPE_IMAGE, ConnectableAttribute.OUTPUT, parent = self)

        self.pixel_size = 10.0
        self.default_sigma = 30.0
        self.fix_sigma = False

        self.reconstructor = Reconstructor()
        self.latest_image = None

        self.original_pixel_size = 100.0
        self.reconstruction_image_size = [1, 1]

        # painting particles
        self.paint_particles = False
        self.paint_by = 0
        self.paint_applied = False

        self.parameter = 0
        self.available_parameters = ["..."]
        self.colour_min = (0.0, 0.0, 0.0)
        self.colour_max = (1.0, 1.0, 1.0)
        self.histogram_values = np.asarray([0, 0]).astype('float32')
        self.histogram_bins = np.asarray([0, 0]).astype('float32')
        self.histogram_initiated = False
        self.min = 0
        self.max = 0
        #
        self.does_profiling_time = False
        self.does_profiling_count = False

        self.auto_render = False
        self.output_mode = 0
        self.OVERRIDE_AUTOCONTRAST = False
        self.OVERRIDE_AUTOCONTRAST_LIMS = (0, 65535)

    def render(self):
        if super().render_start():
            self.reconstruction_in.render_start()
            self.image_out.render_start()
            self.reconstruction_in.render_end()
            self.image_out.render_end()

            if self.reconstruction_in.check_connect_event() or self.reconstruction_in.check_disconnect_event():
                self.on_gain_focus()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            imgui.push_item_width(90)
            imgui.text("Pixel size:")
            imgui.same_line()
            _c, self.pixel_size = imgui.input_float("##Pixel size", self.pixel_size, 0.0, 0.0, format='%.2f (nm)')
            pxs_changed = _c
            self.any_change = _c or self.any_change
            imgui.text(f"Final image size: {self.reconstruction_image_size[0]} x {self.reconstruction_image_size[1]} px")

            # Colourize optionsx
            _c, self.paint_particles = imgui.checkbox("Colourize particles", self.paint_particles)
            self.BLOCK_AUTOCONTRAST_ON_NEW_FRAME = self.paint_particles
            if _c and not self.paint_particles:
                self.paint_by = 0
                self.reconstructor.set_lut(0)
            if self.paint_particles:
                _cw = imgui.get_content_region_available_width()
                imgui.set_next_item_width(_cw - 80)
                _c, self.parameter = imgui.combo("Parameter", self.parameter, self.available_parameters)
                if _c:
                    self.get_histogram_values()
                    self.min = self.histogram_bins[0]
                    self.max = self.histogram_bins[-1]
                    self.paint_applied = False
                imgui.set_next_item_width(_cw - 80)
                _c, self.paint_by = imgui.combo("LUT", self.paint_by, settings.lut_names)
                if _c:
                    self.reconstructor.set_lut(self.paint_by)
                    self.paint_applied = False
                if True:
                    imgui.plot_histogram("##hist", self.histogram_values, graph_size=(_cw, 80.0))
                    _l = self.histogram_bins[0]
                    _h = self.histogram_bins[-1]
                    _max = self.max
                    _min = self.min
                    _uv_left = 0.5
                    _uv_right = 0.5
                    if _max != _min:
                        _uv_left = 1.0 + (_l - _max) / (_max - _min)
                        _uv_right = 1.0 + (_h - _max) / (_max - _min)
                    imgui.image(self.reconstructor.lut_texture.renderer_id, _cw, 10.0, (_uv_left, 0.5),
                                (_uv_right, 0.5), border_color=(0.0, 0.0, 0.0, 1.0))
                    imgui.push_item_width(_cw)
                    _c, self.min = imgui.slider_float("##min", self.min, self.histogram_bins[0], self.histogram_bins[1],format='min: %.1f')
                    if _c: self.paint_applied = False
                    _c, self.max = imgui.slider_float("##max", self.max, self.histogram_bins[0], self.histogram_bins[1],format='max: %.1f')
                    if _c: self.paint_applied = False
                    imgui.pop_item_width()

                # TODO: histogram

            if self.auto_render:
                if self.any_change:
                    self.build_reconstruction()
            else:
                _cw = imgui.get_content_region_available_width()
                imgui.new_line()
                imgui.same_line(spacing = _cw / 2 - 70 / 2)
                if imgui.button("Render", 70, 30):
                    self.build_reconstruction()

            if pxs_changed:
                roi = self.get_particle_data().reconstruction_roi
                img_width = int((roi[3] - roi[1]) / self.pixel_size)
                img_height = int((roi[2] - roi[0]) / self.pixel_size)
                self.reconstruction_image_size = (img_width, img_height)

            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()

            super().render_end()


    def render_advanced(self):
        _c, self.fix_sigma = imgui.checkbox("Force uncertainty", self.fix_sigma)

        if self.fix_sigma:
            imgui.same_line(spacing=10)
            imgui.push_item_width(50)
            _, self.default_sigma = imgui.input_float(" nm", self.default_sigma, 0, 0, format="%.1f")
            imgui.pop_item_width()

        _c, self.output_mode = imgui.combo("Output mode", self.output_mode, ["float", "uint16"])
        if _c:
            if self.output_mode == 0:
                pass
                self.reconstructor.set_mode("float")
            else:
                pass
                self.reconstructor.set_mode("ui16")

        # if imgui.button("Recompile shader"):
        #     self.reconstructor.recompile_shader()

    def get_histogram_values(self):
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            particledata = datasource.get_particle_data()
            if self.available_parameters[self.parameter] in list(particledata.histogram_counts.keys()):
                self.histogram_values = particledata.histogram_counts[self.available_parameters[self.parameter]]
                self.histogram_bins = particledata.histogram_bins[self.available_parameters[self.parameter]]

    def init_histogram_values(self):
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            self.histogram_initiated = True
            particledata = datasource.get_particle_data()
            if particledata:
                if particledata.baked:
                    self.available_parameters = list(particledata.histogram_counts.keys())
                    self.parameter = min([1, len(self.available_parameters)])
                    self.histogram_values = particledata.histogram_counts[self.available_parameters[self.parameter]]
                    self.histogram_bins = particledata.histogram_bins[self.available_parameters[self.parameter]]
                    self.min = self.histogram_bins[0]
                    self.max = self.histogram_bins[-1]


    def apply_paint(self, particle_data):
        if cfg.profiling:
            time_start = time.time()

        values = particle_data.parameter[self.available_parameters[self.parameter]]
        i = 0
        for particle in particle_data.particles:
            particle.colour_idx = np.min([np.max([0.0, (values[i] - self.min) / (self.max - self.min)]), 1.0])
            i += 1
        self.reconstructor.colours_set = False
        if cfg.profiling:
            self.profiler_time += time.time() - time_start

    def build_reconstruction(self):
        try:
            self.original_pixel_size = Node.get_source_load_data_node(self).dataset.pixel_size
            self.reconstructor.set_pixel_size(self.pixel_size)
            self.reconstructor.set_image_size(self.reconstruction_image_size)
            datasource = self.reconstruction_in.get_incoming_node()
            if datasource:
                particle_data = datasource.get_particle_data()
                self.reconstructor.set_particle_data(particle_data)
                self.reconstructor.set_camera_origin([-particle_data.reconstruction_roi[0] / self.pixel_size, -particle_data.reconstruction_roi[1] / self.pixel_size])

                ## Apply colours
                ## TODO: add option to output a monochrome channel
                if self.paint_particles:
                    if not self.paint_applied:
                        self.apply_paint(particle_data)
                        self.paint_applied = True
                        self.reconstructor.colours_set = False
                else:
                    if not self.paint_applied:
                        for particle in particle_data.particles:
                            particle.colour_idx = 1.0
                        self.paint_applied = True
                        self.reconstructor.colours_set = False
                if self.reconstructor.particle_data.empty:
                    return None
                else:
                    self.latest_image = self.reconstructor.render(fixed_uncertainty=(self.default_sigma if self.fix_sigma else None))
                    if not self.paint_particles:
                        self.latest_image = self.latest_image[:, :, 0]  # single channel only for non-painted particles
                    self.any_change = True
                    if self.paint_particles:
                        self.OVERRIDE_AUTOCONTRAST = True
                        clims = self.compute_contrast_lims(self.latest_image)
                        self.OVERRIDE_AUTOCONTRAST_LIMS = (clims[0], clims[1])
            else:
                self.latest_image = None
        except Exception as e:
            cfg.set_error(e, "Error building reconstruction.\n"+str(e))

    @staticmethod
    def compute_contrast_lims(rgb_image):
        max_channel = np.unravel_index(np.argmax(rgb_image), rgb_image.shape)
        img_sorted = np.sort(rgb_image[:, :, max_channel].flatten())
        n = img_sorted.shape[0] * img_sorted.shape[1]
        cmin = img_sorted[int(settings.autocontrast_saturation / 100.0 * n)]
        cmax = img_sorted[int((1.0 - settings.autocontrast_saturation / 100.0) * n)]
        return cmin, cmax

    def get_image_impl(self, idx=None):
        if self.latest_image is not None:
            img_wrapper = Frame("super-resolution reconstruction virtual frame")
            img_wrapper.data = self.latest_image
            img_wrapper.pixel_size = self.pixel_size
            return img_wrapper
        else:
            return None

    def get_particle_data_impl(self):
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            return datasource.get_particle_data()
        else:
            return ParticleData()

    def on_gain_focus(self):
        self.init_histogram_values()
        roi = None
        try:
            roi = self.get_particle_data().reconstruction_roi
        except Exception as e:
            cfg.set_error(e, f"Couldn't get particle data in {self.title} node.")
        if roi is not None:
            img_width = int((roi[2] - roi[0]) / self.pixel_size)  # TODO
            img_height = int((roi[3] - roi[1]) / self.pixel_size)
            self.reconstruction_image_size = (img_width, img_height)

    def pre_save_impl(self):
        cfg.pickle_temp["latest_image"] = self.latest_image
        self.latest_image = None

    def post_save_impl(self):
        self.latest_image = cfg.pickle_temp["latest_image"]




from scNodes.core import settings
from scNodes.core import config as cfg
from scNodes.core.opengl_classes import *
from PIL import Image
import mrcfile
from copy import copy, deepcopy
import datetime


class CLEMFrame:
    idgen = count(0)
    HISTOGRAM_BINS = 40

    def __init__(self, img_array):
        """Grayscale images only - img_array must be a 2D np.ndarray"""
        # data
        uid_counter = next(CLEMFrame.idgen)
        self.uid = int(datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')+"000") + uid_counter
        self.data = img_array
        self.is_rgb = False
        if len(self.data.shape) == 2:
            self.height, self.width = self.data.shape
        elif len(self.data.shape) == 3:
            self.height = self.data.shape[0]
            self.width = self.data.shape[1]
            self.is_rgb = True
            self.data = self.data[:, :, 0:3]
        else:
            raise Exception("CLEMFrame not able to import image data with dimensions other than (XY) or (XYC). How did you manage..?")
        self.children = list()
        self.parent = None
        self.title = "Frame "+str(uid_counter)
        self.path = None
        self.has_slices = False
        self.n_slices = 1
        self.current_slice = 0
        self.extension = ""
        # transform parameters
        self.pixel_size = cfg.ce_default_pixel_size  # pixel size in nm
        self.pivot_point = np.zeros(2)  # pivot point for rotation and scaling of this particular image. can be moved by the user. In _local coordinates_, i.e. relative to where the frame itself is positioned.
        self.transform = Transform()
        self.flip_h = False
        self.flip_v = False
        # visuals
        self.binning = 1
        self.blend_mode = 0
        self.lut = 1
        self.lut_clamp_mode = 0
        self.colour = (1.0, 1.0, 1.0, 1.0)
        self.alpha = 1.0
        self.contrast_lims = [0.0, 65535.0]
        self.compute_autocontrast()
        self.hist_bins = list()
        self.hist_vals = list()
        self.rgb_hist_vals = list()
        self.rgb_contrast_lims = [0, 255, 0, 255, 0, 255]
        self.compute_histogram()
        self.hide = False
        self.interpolate = False  # False for pixelated, True for interpolated

        # aux
        self.corner_positions_local = [[0, 0], [0, 0], [0, 0], [0, 0]]
        self.corner_positions = [[0, 0], [0, 0], [0, 0], [0, 0]]

        # opengl
        self.texture = None
        self.lut_texture = None
        self.quad_va = None
        self.vertex_positions = None
        self.border_va = None
        self.setup_opengl_objects()

    def update(self):
        self.update_image_texture()

    def duplicate(self):
        my_copy = deepcopy(self)
        uid_counter = next(CLEMFrame.idgen)
        my_copy.uid = int(datetime.datetime.now().strftime('%Y%m%d%H%M%S%f') + "000") + uid_counter
        my_copy.setup_opengl_objects()
        return my_copy

    def flip(self, horizontally=True):
        if horizontally:
            self.flip_h = not self.flip_h
            self.data = np.flip(self.data, axis=1)
        else:
            self.flip_v = not self.flip_v
            self.data = np.flip(self.data, axis=0)
        self.update_image_texture(compute_histogram=False)

    def setup_opengl_objects(self):
        if self.is_rgb:
            self.texture = Texture(format="rgb32f")
            self.texture.update(self.data.astype(np.float32))
        else:
            self.texture = Texture(format="r32f")
            self.texture.update(self.data.astype(np.float32))
        self.lut_texture = Texture(format="rgba32f")
        self.quad_va = VertexArray()
        self.vertex_positions = list()
        self.border_va = VertexArray(attribute_format="xy")
        self.update_lut()
        self.generate_va()

    def update_image_texture(self, compute_histogram=True):
        self.texture.update(self.data.astype(np.float32))
        if not self.is_rgb and compute_histogram:
            self.compute_histogram()

    def set_slice(self, requested_slice):
        if not self.has_slices:
            return
        requested_slice = min([max([requested_slice, 0]), self.n_slices - 1])
        if requested_slice == self.current_slice:
            return
        # else, load slice and update texture.
        self.current_slice = requested_slice
        if self.extension == ".tiff" or self.extension == ".tif":
            data = Image.open(self.path)
            data.seek(self.current_slice)
            self.data = np.asarray(data)
            if cfg.ce_flip_on_load:
                self.data = np.flip(self.data, axis=1)
            if self.flip_h:
                self.data = np.flip(self.data, axis=1)
            if self.flip_v:
                self.data = np.flip(self.data, axis=0)
            self.update_image_texture()
        elif self.extension == ".mrc":
            mrc = mrcfile.mmap(self.path, mode="r")
            self.n_slices = mrc.data.shape[0]
            self.current_slice = min([self.current_slice, self.n_slices - 1])
            if cfg.ce_flip_mrc_on_load:
                self.data = mrc.data[:, self.current_slice, :]
            else:
                self.data = mrc.data[self.current_slice, :, :]
            if self.flip_h:
                self.data = np.flip(self.data, axis=1)
            if self.flip_v:
                self.data = np.flip(self.data, axis=0)
            self.update_image_texture()

    def translate(self, translation):
        self.pivot_point[0] += translation[0]
        self.pivot_point[1] += translation[1]
        self.transform.translation[0] += translation[0]
        self.transform.translation[1] += translation[1]

    def pivoted_rotation(self, pivot, angle):
        self.transform.rotation += angle
        p = np.matrix([self.transform.translation[0] - pivot[0], self.transform.translation[1] - pivot[1]]).T
        rotation_mat = np.identity(2)
        _cos = np.cos(angle / 180.0 * np.pi)
        _sin = np.sin(angle / 180.0 * np.pi)
        rotation_mat[0, 0] = _cos
        rotation_mat[1, 0] = _sin
        rotation_mat[0, 1] = -_sin
        rotation_mat[1, 1] = _cos
        delta_translation = rotation_mat * p - p
        delta_translation = [float(delta_translation[0]), float(delta_translation[1])]
        self.transform.translation[0] += delta_translation[0]
        self.transform.translation[1] += delta_translation[1]

    def pivoted_scale(self, pivot, scale):
        offset = [pivot[0] - self.transform.translation[0], pivot[1] - self.transform.translation[1]]
        self.transform.translation[0] += offset[0] * (1.0 - scale)
        self.transform.translation[1] += offset[1] * (1.0 - scale)
        self.pixel_size *= scale

    def update_model_matrix(self):
        self.transform.scale = self.pixel_size
        self.transform.compute_matrix()
        for i in range(4):
            vec = np.matrix([*self.corner_positions_local[i], 0.0, 1.0]).T
            corner_pos = (self.transform.matrix * vec)[0:2]
            self.corner_positions[i] = [float(corner_pos[0]), float(corner_pos[1])]

    def world_to_pixel_coordinate(self, world_coordinate):
        vec = np.matrix([world_coordinate[0], world_coordinate[1], 0.0, 1.0]).T
        invmat = np.linalg.inv(self.transform.matrix)
        out_vec = invmat * vec
        va_coordinate = [out_vec[0, 0], out_vec[1, 0]]
        pixel_coordinate = [int(out_vec[0,0] + self.width / 2), int(out_vec[1,0] + self.height / 2)]
        return pixel_coordinate, va_coordinate

    def unparent(self):
        if self.parent is not None:
            self.parent.children.remove(self)
        self.parent = None

    def parent_to(self, parent):
        if parent is None:
            self.unparent()
            return
        # check that parent isn't any child of self
        _parent = parent
        inheritance_loop = False
        while _parent is not None:
            if _parent == self:
                inheritance_loop = True
                break
            _parent = _parent.parent
        if inheritance_loop:
            return

        # remove self from parent's children list, set parent, and edit parent's children
        self.unparent()
        self.parent = parent
        parent.children.append(self)

    def list_all_children(self, include_self=False):
        """returns a list with all of this frame's children, + children's children, + etc."""
        def get_children(frame):
            children = list()
            for child in frame.children:
                children.append(child)
                children += get_children(child)
            return children
        all_children = get_children(self)
        if include_self:
            all_children = [self] + all_children
        return all_children

    def toggle_interpolation(self):
        self.interpolate = not self.interpolate
        if self.interpolate:
            self.texture.set_linear_interpolation()
        else:
            self.texture.set_no_interpolation()

    def force_lut_grayscale(self):
        original_lut = copy(self.lut)
        self.lut = 1
        self.update_lut()
        self.lut = original_lut

    def update_lut(self):
        if self.lut > 0:
            lut_array = np.asarray(settings.luts[settings.lut_names[self.lut - 1]])
            self.colour = (lut_array[-1, 0], lut_array[-1, 1], lut_array[-1, 2], 1.0)
        else:
            lut_array = np.asarray(settings.luts[settings.lut_names[0]]) * np.asarray(self.colour[0:3])
        if lut_array.shape[1] == 3:
            lut_array = np.reshape(lut_array, (1, lut_array.shape[0], 3))
        # Add alpha
        lut_array_rgba = np.ones((1, lut_array.shape[1], 4))
        lut_array_rgba[:, :, 0:3] = lut_array
        if self.lut_clamp_mode == 1:
            lut_array_rgba[0, 0, 3] = 0.0
            lut_array_rgba[0, -1, 3] = 0.0
        elif self.lut_clamp_mode == 2:
            lut_array_rgba[0, 0, 3] = 0.0
        self.lut_texture.update(lut_array_rgba)

    def compute_autocontrast(self):
        subsample = self.data[::settings.autocontrast_subsample, ::settings.autocontrast_subsample]
        n = subsample.shape[0] * subsample.shape[1]
        sorted_pixelvals = np.sort(subsample.flatten())
        self.contrast_lims[0] = sorted_pixelvals[int(settings.autocontrast_saturation / 100.0 * n)]
        self.contrast_lims[1] = sorted_pixelvals[int((1.0 - settings.autocontrast_saturation / 100.0) * n)]

    def compute_histogram(self):
        if not self.is_rgb:
            # ignore very bright pixels
            mean = np.mean(self.data)
            std = np.std(self.data)
            self.hist_vals, self.hist_bins = np.histogram(self.data[self.data < (mean + 10 * std)], bins=CLEMFrame.HISTOGRAM_BINS)

            self.hist_vals = self.hist_vals.astype('float32')
            self.hist_bins = self.hist_bins.astype('float32')
            self.hist_vals = np.delete(self.hist_vals, 0)
            self.hist_bins = np.delete(self.hist_bins, 0)
            self.hist_vals = np.log(self.hist_vals + 1)
        else:
            for i in range(3):
                vals, _ = np.histogram(self.data[:, :, i], bins=CLEMFrame.HISTOGRAM_BINS)
                vals = vals.astype('float32')
                vals = np.delete(vals, 0)
                self.rgb_hist_vals.append(vals)

    def generate_va(self):
        # set up the quad vertex array
        w, h = self.width * 0.5, self.height * 0.5
        vertex_attributes = [-w, h, 1.0, 0.0, 1.0,
                             -w, -h, 1.0, 0.0, 0.0,
                             w, -h, 1.0, 1.0, 0.0,
                             w, h, 1.0, 1.0, 1.0]
        self.corner_positions_local = [[-w, h], [-w, -h], [w, -h], [w, h]]
        indices = [0, 1, 2, 2, 0, 3]

        ## TODO: tesselate
        self.quad_va.update(VertexBuffer(vertex_attributes), IndexBuffer(indices))

        # set up the border vertex array
        vertex_attributes = [-w, h,
                             w, h,
                             w, -h,
                             -w, -h]
        indices = [0, 1, 1, 2, 2, 3, 3, 0]
        self.border_va.update(VertexBuffer(vertex_attributes), IndexBuffer(indices))

    def move_to_front(self):
        cfg.ce_frames.insert(0, cfg.ce_frames.pop(cfg.ce_frames.index(self)))

    def move_to_back(self):
        cfg.ce_frames.append(cfg.ce_frames.pop(cfg.ce_frames.index(self)))

    def move_backwards(self):
        idx = cfg.ce_frames.index(self)
        if idx < (len(cfg.ce_frames) - 1):
            cfg.ce_frames[idx], cfg.ce_frames[idx + 1] = cfg.ce_frames[idx + 1], cfg.ce_frames[idx]

    def move_forwards(self):
        idx = cfg.ce_frames.index(self)
        if idx > 0:
            cfg.ce_frames[idx], cfg.ce_frames[idx - 1] = cfg.ce_frames[idx - 1], cfg.ce_frames[idx]

    def __eq__(self, other):
        if isinstance(other, CLEMFrame):
            return self.uid == other.uid
        return False

    def __str__(self):
        return f"CLEMFrame with id {self.uid}."


class Transform:
    def __init__(self):
        self.translation = np.array([0.0, 0.0])
        self.rotation = 0.0
        self.scale = 1.0
        self.matrix = np.identity(4)

    def compute_matrix(self):
        scale_mat = np.identity(4) * self.scale
        scale_mat[3, 3] = 1

        rotation_mat = np.identity(4)
        _cos = np.cos(self.rotation / 180.0 * np.pi)
        _sin = np.sin(self.rotation / 180.0 * np.pi)
        rotation_mat[0, 0] = _cos
        rotation_mat[1, 0] = _sin
        rotation_mat[0, 1] = -_sin
        rotation_mat[1, 1] = _cos

        translation_mat = np.identity(4)
        translation_mat[0, 3] = self.translation[0]
        translation_mat[1, 3] = self.translation[1]

        self.matrix = np.matmul(translation_mat, np.matmul(rotation_mat, scale_mat))

    def __add__(self, other):
        out = Transform()
        out.translation[0] = self.translation[0] + other.translation[0]
        out.translation[1] = self.translation[1] + other.translation[1]
        out.rotation = self.rotation + other.rotation
        return out

    def __str__(self):
        return f"Transform with translation = {self.translation[0], self.translation[1]}, scale = {self.scale}, rotation = {self.rotation}"

    def __sub__(self, other):
        out = Transform()
        out.translation[0] = self.translation[0] - other.translation[0]
        out.translation[1] = self.translation[1] - other.translation[1]
        out.rotation = self.rotation - other.rotation
        return out



from scNodes.core.node import *
from tkinter import filedialog
from scNodes.core.util import get_filetype


def create():
    return LoadDataNode()


class LoadDataNode(Node):
    title = "Import dataset"
    group = "Data IO"

    colour = (84 / 255, 77 / 255, 222 / 255, 1.0)

    sortid = 0

    def __init__(self):
        super().__init__()
        self.size = 200

        # Set up connectable attributes
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent=self)

        # flags
        self.NODE_IS_DATA_SOURCE = True

        # Set up node-specific vars
        self.dataset = Dataset()
        self.path = ""
        self.pixel_size = 64.0
        self.load_on_the_fly = True
        self.done_loading = False
        self.to_load_idx = 0
        self.n_to_load = 1

        self.file_filter_positive_raw = ""
        self.file_filter_negative_raw = ""

    def render(self):
        if super().render_start():
            self.dataset_out.render_start()
            self.dataset_out.render_end()
            imgui.spacing()
            imgui.separator()
            imgui.spacing()
            imgui.text("Select source file")
            imgui.push_item_width(150)
            _, self.path = imgui.input_text("##intxt", self.path, 256, imgui.INPUT_TEXT_ALWAYS_OVERWRITE)
            imgui.pop_item_width()
            imgui.same_line()
            if imgui.button("...", 26, 19):
                selected_file = filedialog.askopenfilename()
                if selected_file is not None:
                    if get_filetype(selected_file) in ['.tiff', '.tif']:
                        self.path = selected_file
                        self.on_select_file()
            imgui.columns(2, border = False)
            imgui.text("frames:")
            imgui.text("image size:")
            imgui.text("pixel size:  ")
            imgui.next_column()
            imgui.new_line()
            imgui.same_line(spacing=3)
            imgui.text(f"{self.dataset.n_frames}")
            imgui.new_line()
            imgui.same_line(spacing=3)
            imgui.text(f"{self.dataset.img_width}x{self.dataset.img_height}")
            imgui.push_item_width(45)
            _c, self.pixel_size = imgui.input_float("##nm", self.pixel_size, 0.0, 0.0, format = "%.1f")
            self.any_change = self.any_change or _c
            if _c:
                self.dataset.pixel_size = self.pixel_size
            imgui.pop_item_width()
            imgui.same_line()
            imgui.text("nm")
            imgui.columns(1)


            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()

            super().render_end()

    def render_advanced(self):
        _, self.load_on_the_fly = imgui.checkbox("Load on the fly", self.load_on_the_fly)
        if not self.load_on_the_fly and not self.done_loading:
            self.progress_bar(min([self.to_load_idx / self.n_to_load]))
            imgui.spacing()
            imgui.spacing()
            imgui.spacing()

        imgui.text("Title must contain:")
        av_width = imgui.get_content_region_available_width()
        imgui.set_next_item_width(av_width)
        _c, self.file_filter_positive_raw = imgui.input_text("##filt_pos", self.file_filter_positive_raw, 1024, imgui.INPUT_TEXT_AUTO_SELECT_ALL)
        Node.tooltip("Enter tags, e.g. 'GFP;RFP', that must be in a filename\n"
                     "in order to retain the frame. Separate tags by a semicolon.\n"
                     "When multiple tags are entered, frames are retained if \n"
                     "any of the tags is in the filename (not necessarily all).\n"
                     "Leave empty to retain all files by default.")
        imgui.text("Title must not contain:")
        imgui.set_next_item_width(av_width)
        _c, self.file_filter_negative_raw = imgui.input_text("##filt_neg", self.file_filter_negative_raw, 1024, imgui.INPUT_TEXT_AUTO_SELECT_ALL)
        Node.tooltip("Enter tags, e.g. 'GFP;RFP', to select frames for deletion.\n"
                     "Separate by a semicolon. When the positive and negative\n"
                     "selection criteria contradict, frames are retained.")
        if imgui.button("Filter", av_width / 2 - 5, 25):
            self.dataset.filter_frames_by_title(self.file_filter_positive_raw, self.file_filter_negative_raw)
            self.any_change = True
        imgui.same_line(spacing=10)
        if imgui.button("Reset", av_width / 2 - 5, 25):
            self.on_select_file()
            self.any_change = True

    def on_select_file(self):
        try:
            self.dataset = Dataset(self.path, self.pixel_size)
            self.n_to_load = self.dataset.n_frames
            self.done_loading = False
            self.to_load_idx = 0
            self.any_change = True
            cfg.image_viewer.center_image_requested = True
            cfg.set_active_node(self)
        except Exception as e:
            cfg.set_error(e, f"Error importing '{self.path}' as tif stack. Are you sure the data is .tif and 1 colour channel?")

    def get_image_impl(self, idx):
        if self.dataset.n_frames > 0:
            retimg = copy.deepcopy(self.dataset.get_indexed_image(idx))
            retimg.pixel_size = self.pixel_size
            retimg.clean()
            return retimg
        else:
            return None

    def on_update(self):
        if not self.load_on_the_fly and not self.done_loading:
            if cfg.profiling:
                time_start = time.time()
                self.profiler_count += 1
            if self.to_load_idx < self.dataset.n_frames:
                self.dataset.get_indexed_image(self.to_load_idx).load()
                self.to_load_idx += 1
            else:
                self.done_loading = True
            if cfg.profiling:
                self.profiler_time += time.time() - time_start

    def pre_save_impl(self):
        cfg.pickle_temp["dataset"] = self.dataset
        self.dataset = Dataset()

    def post_save_impl(self):
        self.dataset = cfg.pickle_temp["dataset"]

    def on_load(self):
        # TODO
        pass




from scNodes.core.node import *

"""
This file can be used as a template to build a custom node. See the user manual for more information.
For a node to be properly registered by the software, the .py file that defines the node must include:
i) a function 'create' that returns an object of type Node
ii) a class that defines said Node object. A new node is easily set up by making a new class, e.g. 'CustomNode,' that inherits from the Node base class.
"""


def create():
    return CustomNode()


class CustomNode(Node):
    title = "Example node"  # The title is the name by which the node appears in the editor.
    group = ["Tutorial", "Custom nodes"]  # In the 'add node' right-click context menu, nodes are listed in groupes. A node can be in multiple groups as in this example. Any group name can be used here - also groups that are not in the default software.
    colour = (1.0, 0.0, 1.0, 1.0)  # this node will be magenta.
    sortid = 0  # upon initializing software, nodes are loaded in order of increasing sortid value. This determines the order with which they are presented in the editor right click context menu.
    enabled = False

    def __init__(self):
        super().__init__()  # In this line the init function of the Node parent class is called (can be found in node.py) - you can ignore it but it must be called.
        self.size = 300  # Set the horizontal size of the node.

        # defining in- and output attributes.
        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.INPUT, parent=self)  # In this example the node has two 'connectable attributes': a dataset input and output.
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.OUTPUT, parent=self)
        # below we define some variables that are used for the example imgui features.
        self.noise_stdev = 1.0
        self.n_button_clicks = 0
        self.stopwatch = False
        self.stopwatch_start = 0
        self.stopwatch_now = 0

    def render(self):
        if super().render_start(): # as in the above __init__ function, the render function must by calling the base class' render_start() function, and end with a matching render_end() - see below.
            self.dataset_in.render_start()  # calling a ConnectableAttribute's render_start() and render_end() handles the rendering and connection logic for that attribute.
            self.dataset_out.render_start()
            self.dataset_in.render_end()
            self.dataset_out.render_end() # callin start first for both attributes and end afterwards makes the attributes appear on the same line / at the same height.

            imgui.spacing()  # add a small vertical whitespace
            imgui.separator()  # draw a line separating the above connectors from the rest of the body of the node. purely visual.
            imgui.spacing()

            # THE BELOW SECTION SHOWCASES SOME COMMON IMGUI FEATURES THAT CAN BE USED. For a detailed guide refer to the documentation of pyimgui: https://pyimgui.readthedocs.io/en/latest/
            imgui.text("This custom node generates noise.")
            imgui.set_next_item_width(150)
            _c, self.noise_stdev = imgui.slider_float("Noise stdev", self.noise_stdev, 0.0, 10.0)
            self.any_change = self.any_change or _c  # Node.any_change is one of various flags; boolean member variables of the Node base class that trigger certain behaviours. See the user manual for more info. This particular flag notifies the image viewer that a change has been made to the settings, and triggers the image viewer to request an updated output image.

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            button_width = 80
            imgui.new_line()
            remaining_window_content_width = imgui.get_window_content_region_width()
            imgui.same_line(position=remaining_window_content_width / 2 - button_width / 2)
            button_clicked = imgui.button("Hover me", width = button_width, height = 20)
            self.tooltip("now click to increase the counter.")
            if button_clicked:
                self.n_button_clicks += 1
            imgui.text(f"The button was clicked {self.n_button_clicks} times.")

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            _changed, self.colour = imgui.color_edit4("Node colour", *self.colour, imgui.COLOR_EDIT_NO_INPUTS)

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()

            super().render_end()

    def render_advanced(self):
        _changed, self.stopwatch = imgui.checkbox("Stopwatch on", self.stopwatch)
        if _changed:
            self.stopwatch_start = datetime.datetime.now()
            self.stopwatch_now = datetime.datetime.now()
        if self.stopwatch:
            delta_time = self.stopwatch_now - self.stopwatch_start
            imgui.text(f"time passed: " + str(delta_time))


    def on_update(self):
        if self.stopwatch:
            self.stopwatch_now = datetime.datetime.now()

    def get_image_impl(self, idx=None):
        data_source = self.dataset_in.get_incoming_node()  # ask the dataset_in attribute for the node that it has got incoming; i.e., the dataset output attribute it is connected to. If it is not connected, a Node of type NullNode is returned - which eveluates to 'False'
        if data_source:
            input_frame = data_source.get_image(idx)  # get the incoming image (a Frame object), and call its .load() method to get the image's pixel data as a numpy array. The .load() method does some internal optimizations to ensure the image is only ever loaded from disk once, but the raw data remains intact as well.
            input_pxd = input_frame.load()
            out_frame = input_frame.clone()
            out_frame.data = np.random.normal(0, self.noise_stdev, input_pxd.shape)
            return out_frame
        else:
            out_frame = Frame(path = "virtual frame")  # create a new Frame object. Since it is not part of a raw data source, give it a fake path 'virtual frame'
            out_frame.data = np.random.normal(0, self.noise_stdev, (512, 512))  # no input image: output a 512 x 512 pixel noise image.
            return out_frame




import imgui.internal

from scNodes.core.ceplugin import *
from scipy.ndimage import gaussian_filter

def create():
    return BlurPlugin()


class BlurPlugin(CEPlugin):
    title = "Blur frames"
    description = "Apply a Gaussian filter to a frame. This plugin changes the pixel data of a frame object.\n" \
                  "Be sure to duplicate the frame first if you want to keep the original frame in the scene"

    def __init__(self):
        self.selected_frame = None
        self.sigma = 1.0  # blur kernel standard deviation in pixels.

    def render(self):
        self.selected_frame = CEPlugin.widget_show_active_frame_title(label="Selected frame:")

        imgui.set_next_item_width(50)
        _, self.sigma = imgui.input_float("Sigma", self.sigma)

        if CEPlugin.widget_centred_button("Blur"):
            self.process_frame()

    def process_frame(self):
        out_frame = self.selected_frame.duplicate()
        img_data = out_frame.data

        if self.selected_frame.is_rgb:
            img_data[:, :, 0] = gaussian_filter(img_data[:, :, 0], self.sigma)
            img_data[:, :, 1] = gaussian_filter(img_data[:, :, 1], self.sigma)
            img_data[:, :, 2] = gaussian_filter(img_data[:, :, 2], self.sigma)
        else:
            img_data[:, :] = gaussian_filter(img_data, self.sigma)
        out_frame.data = img_data

        cfg.ce_frames.append(out_frame)
        cfg.ce_frames.remove(self.selected_frame)
        out_frame.update()


import os, sys
directory = os.path.join(os.path.dirname(__file__))
directory = directory[:directory.rfind("\\")]
sys.path.append(directory)

if __name__ == "__main__":
    if 'install' in sys.argv:
        from scNodes.install import install
        install()
    else:
        from scNodes.core.window import *
        from scNodes.core.nodeeditor import *
        from scNodes.core.correlation_editor import *
        from scNodes.core.imageviewer import *
        from scNodes.core.reconstruction import *
        from joblib.externals.loky import set_loky_pickler
        set_loky_pickler("cloudpickle")

        cfg.root = os.path.join(os.path.dirname(__file__)).replace("\\", "/")+"/"
        if not glfw.init():
            raise Exception("Could not initialize GLFW library!")


        # Init the main window, its imgui context, and a glfw rendering impl.
        main_window = Window(cfg.window_width, cfg.window_height, settings.ne_window_title)
        main_window.set_callbacks()
        main_window_imgui_context = imgui.create_context()
        main_window_imgui_glfw_implementation = GlfwRenderer(main_window.glfw_window)
        main_window.set_mouse_callbacks()
        main_window.set_window_callbacks()
        node_editor = NodeEditor(main_window, main_window_imgui_context, main_window_imgui_glfw_implementation)
        correlation_editor = CorrelationEditor(main_window, main_window_imgui_context, main_window_imgui_glfw_implementation)


        image_viewer_window_imgui_context = imgui.create_context(main_window_imgui_glfw_implementation.io.fonts)
        image_viewer_window = Window(cfg.iv_window_width, cfg.iv_window_height, settings.iv_window_title)
        image_viewer_window_glfw_implementation = GlfwRenderer(image_viewer_window.glfw_window)
        image_viewer = ImageViewer(image_viewer_window, image_viewer_window_imgui_context, image_viewer_window_glfw_implementation)
        image_viewer_window.set_callbacks()
        image_viewer_window.set_window_callbacks()
        node_editor.delete_temporary_files()
        cfg.node_editor = node_editor
        cfg.correlation_editor = correlation_editor
        cfg.image_viewer = image_viewer


        try:
            while not glfw.window_should_close(main_window.glfw_window):
                if not (main_window.focused or image_viewer_window.focused):
                    glfw.poll_events()

                if cfg.active_editor == 0:
                    node_editor.on_update()
                    node_editor.end_frame()
                else:
                    correlation_editor.on_update()
                    correlation_editor.end_frame()
                image_viewer.window.make_current()
                image_viewer.on_update()
                image_viewer.end_frame()
        except Exception as e:
            cfg.set_error(e, "CRITICAL ERROR IN SCNODES MAIN\n")
            cfg.write_to_log(cfg.error_msg)
        finally:
            node_editor.delete_temporary_files()






import numpy as np
import pygpufit.gpufit as gf
from scNodes.core.reconstruction import Particle



def frame_to_particles(frame, initial_sigma=2.0, method=0, crop_radius=4, constraints=(-1, -1, -1, -1, -1, -1, -1, -1, -1, -1)):
    def get_background_stdev(flat_roi, fitted_params):
        gauss = np.empty_like(flat_roi)
        k = int(np.sqrt(gauss.shape[0])) // 2
        denom = 2 * fitted_params[3] ** 2
        i = 0
        for x in range(-k, k + 1):
            for y in range(-k, k + 1):
                gauss[i] = np.exp(-((fitted_params[2] - x) ** 2 + (fitted_params[1] - y) ** 2) / denom)
                i += 1
        gauss = gauss * fitted_params[0] / (np.pi * denom) + fitted_params[4]
        background = flat_roi - gauss
        return np.std(background)

    if len(frame.maxima) == 0:
        return list()
    pxd = frame.load()
    width, height = pxd.shape

    # Filter maxima and convert to floored int.
    xy = list()
    n_particles = 0
    for i in range(frame.maxima.shape[0]):
        if (0 + crop_radius) < frame.maxima[i, 0] < (width - crop_radius - 1) and (0 + crop_radius) < frame.maxima[i, 1] < (height - crop_radius - 1):
            x, y = np.floor(frame.maxima[i, :]).astype(int)
            xy.append([x, y])
            n_particles += 1

    # Prepare data for gpufit
    data = np.empty((n_particles, (crop_radius * 2 + 1)**2), dtype=np.float32)
    params = np.empty((n_particles, 5), dtype=np.float32)
    for i in range(n_particles):
        x, y = xy[i]
        data[i, :] = pxd[x - crop_radius:x + crop_radius + 1, y - crop_radius:y + crop_radius + 1].flatten()
        initial_offset = data[i, :].min()
        initial_intensity = pxd[x, y] - initial_offset
        params[i, :] = [initial_intensity, crop_radius, crop_radius, initial_sigma, initial_offset]

    estimator = gf.EstimatorID.LSE if method == 0 else gf.EstimatorID.MLE

    # Set up constraints
    constraint_type, constraint_values = parse_constraints(constraints, n_particles)
    parameters, states, chi_squares, number_iterations, execution_time = gf.fit_constrained(data, None,
                                                                                            gf.ModelID.GAUSS_2D, params,
                                                                                            estimator_id=estimator,
                                                                                            max_number_iterations=100,
                                                                                            constraint_types=constraint_type,
                                                                                            constraints=constraint_values)
    # TODO: 3d fitting.
    xy = np.asarray(xy)
    parameters[:, 0] *= 2 * np.pi * parameters[:, 3]**2  # scaling from (maximum value of gaussian) to (number of photons)
    parameters[:, 1:3] -= crop_radius
    if np.sum(states == 0) == 0:
        return list()

    # print('ratio converged         {:6.2f} %'.format(np.sum(states == 0) / n_particles * 100))
    # print('ratio max it. exceeded  {:6.2f} %'.format(np.sum(states == 1) / n_particles * 100))
    # print('ratio singular hessian  {:6.2f} %'.format(np.sum(states == 2) / n_particles * 100))
    # print('ratio neg curvature MLE {:6.2f} %'.format(np.sum(states == 3) / n_particles * 100))
    # print('ratio gpu not read      {:6.2f} %'.format(np.sum(states == 4) / n_particles * 100))

    background_stdev = np.empty(n_particles)
    for i in range(n_particles):
        background_stdev[i] = get_background_stdev(data[i, :], parameters[i, :])

    parameters[:, 1] += xy[:, 1]  # offsetting back into image coordinates (rather than crop coordinates)
    parameters[:, 2] += xy[:, 0]  # offsetting back into image coordinates
    particles = list()
    converged = states == 0
    for i in range(n_particles):
        if converged[i] and parameters[i, 3] != constraint_values[0, 7]:
            particles.append(Particle(frame=frame.framenr, x=parameters[i, 1], y=parameters[i, 2], sigma=parameters[i, 3], intensity=parameters[i, 0], offset=parameters[i, 4], bkgstd=background_stdev[i]))

    return particles


constraint_type_dict = dict()
constraint_type_dict[(True, True)] = gf.ConstraintType.FREE
constraint_type_dict[(False, True)] = gf.ConstraintType.LOWER
constraint_type_dict[(True, False)] = gf.ConstraintType.UPPER
constraint_type_dict[(False, False)] = gf.ConstraintType.LOWER_UPPER


def parse_constraints(constraints, n_particles):
    c_type_intensity = constraint_type_dict[(constraints[0] == -1.0, constraints[1] == -1.0)]
    c_type_sigma = constraint_type_dict[(constraints[6] == -1.0, constraints[7] == -1.0)]
    c_type_offset = constraint_type_dict[(constraints[8] == -1.0, constraints[9] == -1.0)]

    constraint_types = np.asarray([c_type_intensity, gf.ConstraintType.FREE, gf.ConstraintType.FREE, c_type_sigma, c_type_offset], dtype=np.int32)
    constraint_values = np.zeros((n_particles, 10), dtype=np.float32)
    constraint_values[:, :] = constraints
    return constraint_types, constraint_values



import imgui
from scNodes.core.clemframe import *
import time


class CEPlugin:
    title = "Unnamed plugin"
    description = "The text entered here is shown as a tooltip when the plugin is\n" \
                  "selected and hovered in the Correlation Editor / Tools menu. "
    enabled = True


    frames_by_title = list()
    frames_by_title_no_rgb = list()
    frames_by_title_rgb = list()
    FLAG_SHOW_LOCATION_PICKER = False
    selected_position = [0, 0]
    selected_frame = None

    def init(self):
        """Only one CEPlugin exists at a time; when selecting a different type tool
        in the Correlation Editor tools menu, an instance of the selected tool
        object is created and used as the active tool. The CETool class is thus
        fairly minimal, but it does provide some useful wrappers for use in your
        own tool implementations (see the static methods). """
        pass

    def render(self):
        """This is where the GUI of a tool is implemented. Processing may also be
        implemented here, or in on_update. In the programme loop, render() is called
        immediately after on_update"""
        pass

    @staticmethod
    def on_update():
        """Do not override."""
        CEPlugin.frames_by_title = list()
        CEPlugin.frames_by_title_no_rgb = list()
        CEPlugin.frames_by_title_rgb = list()
        for f in cfg.ce_frames:
            CEPlugin.frames_by_title.append(f.title)
            if not f.is_rgb:
                CEPlugin.frames_by_title_no_rgb.append(f.title)
            else:
                CEPlugin.frames_by_title_rgb.append(f.title)
        CEPlugin.selected_position = copy(cfg.ce_selected_position)
        CEPlugin.selected_frame = cfg.ce_active_frame

    @staticmethod
    def widget_show_active_frame_title(label=None):
        if label is not None:
            imgui.text(label)
        if CEPlugin.selected_frame is not None:
            imgui.text(CEPlugin.selected_frame.title)
            return CEPlugin.selected_frame
        else:
            imgui.text("No frame selected.")
            return None

    @staticmethod
    def widget_select_frame_no_rgb(label, current_frame):
        """Wrapper for an imgui combo (drop-down) menu with all of the current
        frames in the Correlation Editor available for selection. Only grayscale
        (no RGB / png) images are offered.
        :param label: The label text displayed alongside the drop down menu
        :param current_frame: CLEMFrame object
        :return: tuple (bool:changed, CLEMFrame: selected frame)
        """
        _cw = imgui.get_content_region_available_width()
        imgui.push_item_width(_cw)
        current_idx = 0
        if current_frame is not None and current_frame.title in CEPlugin.frames_by_title_no_rgb:
            current_idx = CEPlugin.frames_by_title_no_rgb.index(current_frame.title)
        imgui.text(label)
        _c, idx = imgui.combo("##" + label, current_idx, CEPlugin.frames_by_title_no_rgb)
        imgui.pop_item_width()
        selected_frame = cfg.ce_frames[idx] if (idx < len(cfg.ce_frames)) else None
        return _c, selected_frame

    @staticmethod
    def widget_select_frame_rgb(label, current_frame):
        """Wrapper for an imgui combo (drop-down) menu with all of the current
        frames in the Correlation Editor available for selection. RGB frames only.
        :param label: The label text displayed alongside the drop down menu
        :param current_frame: CLEMFrame object
        :return: tuple (bool:changed, CLEMFrame: selected frame)
        """
        _cw = imgui.get_content_region_available_width()
        imgui.push_item_width(_cw)
        current_idx = 0
        if current_frame is not None and current_frame.title in CEPlugin.frames_by_title_rgb:
            current_idx = CEPlugin.frames_by_title_rgb.index(current_frame.title)
        imgui.text(label)
        _c, idx = imgui.combo("##" + label, current_idx, CEPlugin.frames_by_title_rgb)
        imgui.pop_item_width()
        selected_frame = cfg.ce_frames[idx] if (idx < len(cfg.ce_frames)) else None
        return _c, selected_frame

    @staticmethod
    def widget_select_frame_any(label, current_frame):
        """Wrapper for an imgui combo (drop-down) menu with all of the current
        frames in the Correlation Editor available for selection.
        :param label: The label text displayed alongside the drop down menu
        :param current_frame: CLEMFrame object
        :return: tuple (bool:changed, CLEMFrame: selected frame)
        """
        _cw = imgui.get_content_region_available_width()
        imgui.push_item_width(_cw)
        current_idx = 0
        if current_frame is not None and current_frame.title in CEPlugin.frames_by_title:
            current_idx = CEPlugin.frames_by_title.index(current_frame.title)
        imgui.text(label)
        _c, idx = imgui.combo("##" + label, current_idx, CEPlugin.frames_by_title)
        imgui.pop_item_width()
        selected_frame = cfg.ce_frames[idx] if (idx < len(cfg.ce_frames)) else None
        return _c, selected_frame

    @staticmethod
    def widget_selected_position():
        _cw = imgui.get_content_region_available_width()
        imgui.text(f"x: {CEPlugin.selected_position[0] / 1000.0:.2f} um")
        imgui.text(f"y: {CEPlugin.selected_position[1] / 1000.0:.2f} um")

    @staticmethod
    def set_selected_position(x, y):
        cfg.ce_selected_position = [x, y]

    @staticmethod
    def widget_centred_button(label, width=None, height=20):
        _cw = imgui.get_content_region_available_width()
        _button_width = width if width else (_cw - 10) / 2
        _button_height = height
        imgui.new_line()
        imgui.same_line(spacing=(_cw - _button_width) / 2.0)
        imgui.push_style_var(imgui.STYLE_FRAME_ROUNDING, cfg.CE_WIDGET_ROUNDING)
        clicked = False
        if imgui.button(label, _button_width, _button_height):
            clicked = True
        imgui.pop_style_var()
        return clicked

    @staticmethod
    def tooltip(text):
        if imgui.is_item_hovered():
            if cfg.TOOLTIP_HOVERED_TIMER == 0.0:
                cfg.TOOLTIP_HOVERED_START_TIME = time.time()
                cfg.TOOLTIP_HOVERED_TIMER = 0.001  # add a fake 1 ms to get out of this if clause
            elif cfg.TOOLTIP_HOVERED_TIMER > cfg.TOOLTIP_APPEAR_DELAY:
                imgui.set_tooltip(text)
            else:
                cfg.TOOLTIP_HOVERED_TIMER = time.time() - cfg.TOOLTIP_HOVERED_START_TIME
        if not imgui.is_any_item_hovered():
            cfg.TOOLTIP_HOVERED_TIMER = 0.0



from numpy.fft import (fftshift, ifftshift, fftn, ifftn, rfftn, irfftn)
import numpy as np
import sys


def _prep_img_and_psf(image, psf):
    """
    From https://github.com/david-hoffman/pyDecon.
    Do basic data checking, convert data to float, normalize psf and make sure
    data are positive.
    """
    image = image.astype(np.float)
    psf = psf.astype(np.float)
    # need to make sure both image and PSF are totally positive.
    image = _ensure_positive(image)
    psf = _ensure_positive(psf)
    # normalize the kernel
    psf /= psf.sum()
    return image, psf


def _ensure_positive(data):
    """
    From https://github.com/david-hoffman/pyDecon.
    Make sure data is positive and has no zeros for numerical stability.
    """
    data = data.copy()
    data[data <= 0] = np.finfo(data.dtype).eps
    return data


def _zero2eps(data):
    """Make sure data is positive and has no zeros."""
    return np.fmax(data, np.finfo(data.dtype).eps)


def zero_pad(image, shape, position='center'):
    """
    From https: // github.com/aboucaud/pypher/blob/master/pypher/pypher.py.
    Extends image to a certain size with zeros.

    Parameters
    ----------
    image : real 2d ndarray
        Input image.
    shape : tuple of list(int)
        Desired output shape of the image.
    position : str, optional
        The position of the input image in the output one:
            * 'corner'
                top-left corner(default)
            * 'center'
                centered

    Returns
    -------
    padded_img: real ndarray
        The zero-padded image.
    """
    shape = np.asarray(shape, dtype=int)
    imshape = np.asarray(image.shape, dtype=int)

    if np.alltrue(imshape == shape):
        return image

    if np.any(shape <= 0):
        raise ValueError("ZERO_PAD: null or negative shape given")

    dshape = shape - imshape
    if np.any(dshape < 0):
        raise ValueError("ZERO_PAD: target size smaller than source one")

    pad_img = np.zeros(shape, dtype=image.dtype)

    idx, idy = np.indices(imshape)

    if position == 'center':
        if np.any(dshape % 2 != 0):
            raise ValueError("ZERO_PAD: source and target shapes "
                             "have different parity.")
        offx, offy = dshape // 2
    else:
        offx, offy = (0, 0)

    pad_img[idx + offx, idy + offy] = image

    return pad_img


def psf2otf(psf, shape):
    """
    From https: // github.com/aboucaud/pypher/blob/master/pypher/pypher.py.
    Convert point-spread function to optical transfer function.
    Compute the Fast Fourier Transform(FFT) of the point-spread
    function(PSF) array and creates the optical transfer function(OTF)
    array that is not influenced by the PSF off-centering.
    By default, the OTF array is the same size as the PSF array.
    To ensure that the OTF is not altered due to PSF off-centering, PSF2OTF
    post-pads the PSF array(down or to the right) with zeros to match
    dimensions specified in OUTSIZE, then circularly shifts the values of
    the PSF array up ( or to the left) until the central pixel reaches (1, 1)
    position.
    Adapted from MATLAB psf2otf function.

    Parameters
    ----------
    psf : ndarray
        PSF array.
    shape : list(int)
        Output shape of the OTF array.

    Returns
    -------
    otf : ndarray
        OTF array.
    """
    if np.all(psf == 0):
        return np.zeros_like(psf)

    inshape = psf.shape
    # Pad the PSF to outsize
    psf = zero_pad(psf, shape, position='corner')

    # Circularly shift OTF so that the 'center' of the PSF is
    # [0,0] element of the array
    for axis, axis_size in enumerate(inshape):
        psf = np.roll(psf, -int(axis_size / 2), axis=axis)

    # Compute the OTF
    otf = np.fft.fft2(psf)

    # Estimate the rough number of operations involved in the FFT
    # and discard the PSF imaginary part if within roundoff error
    # roundoff error  = machine epsilon = sys.float_info.epsilon
    # or np.finfo().eps
    n_ops = np.sum(psf.size * np.log2(psf.shape))
    otf = np.real_if_close(otf, tol=n_ops)

    return otf


def otf2psf(otf, shape):
    """
    Convert optical transfer function (OTF) to point-spread function (PSF).
    Compute the Inverse Fast Fourier Transform (ifft) of the OTF array and
    creates the PSF array that is not influenced by the OTF off-centering.
    By default, the PSF array is the same size as the OTF array.
    Adapted from MATLAB otf2psf function.

    Parameters
    ----------
    otf : ndarray
        OTF array.
    shape : list (int)
        Output shape of the OTF array.

    Returns
    -------
    psf : ndarray
        PSF array.
    """
    if np.all(otf == 0):
        return np.zeros_like(otf)

    inshape = otf.shape

    # Compute the PSF
    psf = np.fft.ifft2(otf)

    # Estimate the rough number of operations involved in the FFT
    # and discard the PSF imaginary part if within roundoff error.
    n_ops = np.sum(otf.size * np.log2(otf.shape))
    psf = np.real_if_close(psf, tol=n_ops)
    psf = np.real(psf)

    # Circularly shift PSF so that the 'center' of the PSF is
    # [0,0] element of the array
    for axis, axis_size in enumerate(shape):
        psf = np.roll(psf, int(axis_size / 2), axis=axis)

    # Crop output array
    psf = psf[0:shape[0], 0:shape[1]]

    return psf


def corelucy(image, h):
    """
    Make core for the LR estimation. Calculates f to produce the next
    iteration array that maximizes the likelihood that the entire suite
    satisfies the Poisson statistics.
    This is a simplified version of MATLAB corelucy function without
    damping, weights and externally defined functions.

    Parameters
    ----------
    image : ndarray
        Input image.
    h : ndarray
        Zero-padded OTF. h should have the same dimensions as image.

    Returns
    -------
    f : ndarray
        LR extimation core.

    References
    ----------
    .. [1] Acceleration of iterative image restoration algorithms, by D.S.C. Biggs
    and M. Andrews, Applied Optics, Vol. 36, No. 8, 1997.
    .. [2] Deconvolutions of Hubble Space Telescope Images and Spectra,
    R.J. Hanisch, R.L. White, and R.L. Gilliland. in "Deconvolution of Images
    and Spectra", Ed. P.A. Jansson, 2nd ed., Academic Press, CA, 1997.
    """
    u_t = image
    reblur = np.real(ifftn(h * fftn(u_t, u_t.shape), u_t.shape))
    reblur = _ensure_positive(reblur)
    im_ratio = image / reblur
    f = fftn(im_ratio)
    return f


def richardson_lucy(image, psf, iterations=10, **kwargs):
    """
    Richardson-Lucy deconvolution. It deconvolves image using maximum
    likelihood algorithm, returning both deblurred image J and a restored
    point-spread function PSF.
    This is a simplified version of MATLAB deconvblind function without
    damping, weights and externally defined functions.

    Parameters
    ----------
    image : ndarray
       Input degraded image.
    psf : ndarray
       The point spread function.
    iterations : int
       Number of iterations.

    Returns
    -------
    P : ndarray
       Restored point-spread function PSF.
    J : ndarray
        Deblurred image.

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.ndimage.gaussian_filter
    >>> image = np.zeros((8,8),dtype=int)
    >>> image[1::2,::2] = 1
    >>> image[::2,1::2] = 1
    >>> psf = np.zeros((7, 7))
    >>> psf[3, 3] = 1
    >>> psf = gaussian_filter(psf, sigma=[7,7])
    >>> new_psf, deconv_im = richardson_lucy(image, psf, 15)

    Notes
    -----
    The quality of the deconvolution result is greatly dependent on the initial
    PSF size instead of the value. We recommend to calibrate PSF of the imaging
    system and use that as the initial PSF guess. Otherwise, generating a PSF
    according to the magnification of the imaging system is an option.

    References
    ----------
    .. [1] http: // en.wikipedia.org/wiki/Richardson % E2 % 80 % 93Lucy_deconvolution
    .. [2] Biggs, D. S. C.; Andrews, M. Acceleration of Iterative Image
    Restoration Algorithms. Applied Optics 1997, 36 (8), 1766.
    """
    # 1. prepare parameters
    image, psf = _prep_img_and_psf(image, psf)
    sizeI, sizePSF = image.shape, psf.shape
    J, P = {}, {}
    J[1], J[2], J[3], J[4] = image, image, 0, np.zeros((np.prod(sizeI), 2))
    P[1], P[2], P[3], P[4] = psf, psf, 0, np.zeros((np.prod(sizePSF), 2))
    WEIGHT = np.ones(image.shape)
    fw = fftn(WEIGHT)

    # 2. L_R iterations
    for k in range(iterations):
        # 2a. make image and PSF predictions for the next iteration
        Y = np.maximum(J[2], 0)
        B = np.maximum(P[2], 0)
        B /= B.sum()
        # 2b. make core for the LR estimation
        H = psf2otf(B, sizeI)
        CC = corelucy(Y, H)
        # 2c. Determine next iteration image & apply positivity constraint
        J[3] = J[2]
        scale = np.real(ifftn(np.multiply(np.conj(H), fw))) + \
                np.sqrt(np.finfo(H.dtype).eps)
        J[2] = np.maximum(np.multiply(image,
                                      np.real(ifftn(np.multiply(np.conj(H), CC)))) / scale, 0)
        J[4] = np.vstack([J[2].T.reshape(-1, ) - Y.T.reshape(-1, ), J[4][:, 1]]).T
        # 2d. Determine next iteration PSF & 
        #     apply positivity constraint + normalization
        P[3] = P[2]
        H = fftn(J[3])
        scale = otf2psf(np.multiply(np.conj(H), fw), sizePSF) + \
                np.sqrt(np.finfo(H.dtype).eps)
        P[2] = np.maximum(np.multiply(B,
                                      otf2psf(np.multiply(np.conj(H), CC), sizePSF)) / scale, 0)
        P[2] /= P[2].sum()
        P[4] = np.vstack([P[2].T.reshape(-1, ) - B.T.reshape(-1, ), P[4][:, 1]]).T
    P, J = P[2], J[2]  # PSF and updated image
    return P, J


def deconvsk(est_psf, input_im, deconv_lambda, deconv_iter):
    """
    Perform serial Richardson-Lucy deconvolution with shrinking PSFs. 
    U = (U**(l/(l-1))) * (U**(l**2/(l-1))) * ... * (U**(l**n/(l-1))).
    The PSF of the imaging system U can be decomposed into a series a 
    smaller (shrinking) PSF U**r where r > 1, and the image can be 
    deconvolved by these PSFs in sequence. 
    In this way, the result is more similar to the input image, so each 
    individual deconvolution step is a lighter deconcolution task.

    Parameters
    ----------
    est_psf : ndarray
        Estimated PSF.
    input_im : ndarray
        Input image that need deconvolution.
    deconv_lambda : float
        Lambda for the exponent between. It is an empirical parameter
        within the range of (1,2).
    deconv_iter : int
        Number of iterations for each deconvolution.

    Returns
    -------
    deconv_im : ndarray
        Deconvoluted image.

    Notes
    -----
    The quality of the deconvolution result is greatly dependent on the initial 
    PSF size instead of the value. We recommend to calibrate PSF of the imaging
    system and use that as the initial PSF guess. Otherwise, generating a PSF 
    according to the magnification of the imaging system is an option. For more
    details on the shrinking kernel deconvolution method, please refer to [1].

    References
    ----------
    .. [1] Xiyu Yi, Sungho Son, Ryoko Ando, Atsushi Miyawaki, and Shimon Weiss, 
    "Moments reconstruction and local dynamic range compression of high order 
    superresolution optical fluctuation imaging," Biomed. Opt. Express 10, 
    2430-2445 (2019).
    """
    xdim, ydim = np.shape(input_im)
    deconv_im = np.append(np.append(input_im, np.fliplr(input_im), axis=1),
                          np.append(np.flipud(input_im), np.rot90(input_im, 2), axis=1), axis=0)
    # Perform mirror extension to the image in order sto surpress ringing
    # artifacts associated with fourier transform due to truncation effect.
    psf0 = est_psf / np.max(est_psf)
    for iter_num in range(deconv_iter):
        alpha = deconv_lambda ** (iter_num + 1) / (deconv_lambda - 1)
        deconv_psf, deconv_im = richardson_lucy(deconv_im, psf0 ** alpha, 1)
        sys.stdout.write('\r')
        sys.stdout.write("[%-20s] %d%%" % ('=' * iter_num,
                                           100 / (deconv_iter - 1) * iter_num))
        sys.stdout.flush()

    deconv_im = deconv_im[0:xdim, 0:ydim]
    return deconv_im


from scNodes.core.ceplugin import *
from copy import deepcopy

def create():
    return BinPlugin()


class BinPlugin(CEPlugin):
    title = "Apply binning"
    description = "Create a binned copy of a frame."

    MODES = ["Average", "Median", "Min", "Max", "Sum"]
    
    def __init__(self):
        self.frame = None
        self.keep_original = False
        self.mode = 0
        self.factor = 2
        
    def render(self):
        imgui.text("Selected frame:")
        self.frame = self.widget_show_active_frame_title()
        imgui.push_item_width(100)
        _, self.mode = imgui.combo("Mode", self.mode, BinPlugin.MODES)
        _, self.factor = imgui.input_int("Factor", self.factor, 0, 0)
        imgui.pop_item_width()
        _, self.keep_original = imgui.checkbox("Keep original", self.keep_original)
        self.factor = max([self.factor, 1])
        if self.widget_centred_button("Convert"):
            if self.frame is not None:
                binned_frame = BinPlugin.bin_clemframe(self.frame, self.factor, self.mode)
                binned_frame.update_lut()
                cfg.ce_frames.append(binned_frame)
                binned_frame.move_to_front()
                binned_frame.transform = deepcopy(self.frame.transform)
                if not self.keep_original:
                    cfg.correlation_editor.delete_frame(self.frame)

    @staticmethod
    def bin_clemframe(frame, factor, mode):
        if frame.is_rgb:
            binned = np.empty((frame.data.shape[0] // factor, frame.data.shape[1] // factor, 3))
            binned[:, :, 0] = BinPlugin.bin_grayscale(frame.data[:, :, 0], factor, mode)
            binned[:, :, 1] = BinPlugin.bin_grayscale(frame.data[:, :, 0], factor, mode)
            binned[:, :, 2] = BinPlugin.bin_grayscale(frame.data[:, :, 0], factor, mode)
        else:
            binned = BinPlugin.bin_grayscale(frame.data, factor, mode)
        outframe = CLEMFrame(binned)
        outframe.pixel_size = frame.pixel_size * factor
        outframe.lut = copy(frame.lut)
        outframe.contrast_lims = copy(frame.contrast_lims)
        outframe.lut_clamp_mode = copy(frame.lut_clamp_mode)
        outframe.pivot_point = copy(frame.pivot_point)
        outframe.title = frame.title + f"_bin{factor}"
        return outframe

    @staticmethod
    def bin_grayscale(array, factor, mode):
        width, height = array.shape
        pxd = array[:factor * (width // factor), :factor * (height // factor)]
        if mode == 0:
            pxd = pxd.reshape((width // factor, factor, height // factor, factor)).mean(3).mean(1)
        elif mode == 1:
            pxd = pxd.reshape((width // factor, factor, height // factor, factor))
            pxd = np.median(pxd, axis=(3, 1))
        elif mode == 2:
            pxd = pxd.reshape((width // factor, factor, height // factor, factor)).min(3).min(1)
        elif mode == 3:
            pxd = pxd.reshape((width // factor, factor, height // factor, factor)).max(3).max(1)
        elif mode == 4:
            pxd = pxd.reshape((width // factor, factor, height // factor, factor)).sum(3).sum(1)
        return pxd


from scNodes.core.node import *


def create():
    return MathNode()


class MathNode(Node):
    title = "Math"
    colour = (143 / 255, 143 / 255, 143 / 255, 1.0)
    group = "Converters"
    sortid = 2000

    OPERATIONS = ["Power", "Log", "Invert", "Threshold", "Add constant", "Multiply"]

    def __init__(self):
        super().__init__()

        self.dataset_in = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.INPUT, self, [ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.TYPE_IMAGE])
        self.dataset_in.colour = ConnectableAttribute.COLOUR[ConnectableAttribute.TYPE_DATASET]
        self.dataset_out = ConnectableAttribute(ConnectableAttribute.TYPE_MULTI, ConnectableAttribute.OUTPUT, self, [ConnectableAttribute.TYPE_DATASET, ConnectableAttribute.TYPE_IMAGE])
        self.dataset_out.colour = ConnectableAttribute.COLOUR[ConnectableAttribute.TYPE_DATASET]

        self.operation = 0

        # operation specific vars
        self.power = 2.0
        self.threshold = 100
        self.threshold_keep_high = True
        self.threshold_fill_value = 0
        self.constant = 100
        self.factor = 2.0

        self.last_image_min = 0
        self.last_image_max = 256

        self.replace_nan_by = 0
        self.replace_inf_by = 0
        self.replace_inf = False

    def render(self):
        if super().render_start():
            self.dataset_in.render_start()
            self.dataset_out.render_start()
            self.dataset_in.render_end()
            self.dataset_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            imgui.text("Operation")
            imgui.set_next_item_width(170)
            _c, self.operation = imgui.combo("##operation", self.operation, MathNode.OPERATIONS)
            self.any_change = self.any_change or _c
            imgui.push_item_width(120)
            if self.operation == 0:
                _c, self.power = imgui.slider_float("power", self.power, -2.0, 2.0, format = "%.1f")
                self.any_change = self.any_change or _c
            elif self.operation == 3:
                _c, self.threshold = imgui.slider_float("threshold", self.threshold, self.last_image_min, self.last_image_max, format = "%.0f")
                self.any_change = self.any_change or _c
                _c, self.threshold_fill_value = imgui.slider_float("fill", self.threshold_fill_value, self.last_image_min, self.last_image_max, format = "%.0f")
                self.any_change = self.any_change or _c
                _c, self.threshold_keep_high = imgui.checkbox("keep low", self.threshold_keep_high)
                Node.tooltip("By default, values lower than the threshold value are replaced.\n"
                             "Check this box to invert the behaviour: values higher than the\n"
                             "threshold are replaced instead.")
                self.any_change = self.any_change or _c
            elif self.operation == 4:
                _c, self.constant = imgui.slider_float("value", self.constant, -1000, 1000, format = "%.0f")
                self.any_change = self.any_change or _c
            elif self.operation == 5:
                _c, self.factor = imgui.slider_float("factor", self.factor, 0.0, 2.0, format = "%.2f")
                self.any_change = self.any_change or _c
            Node.tooltip("Use ctrl + click on a slider to override it and manually enter a value.")

            imgui.spacing()
            header_expanded, _ = imgui.collapsing_header("Advanced", None)
            if header_expanded:
                self.render_advanced()
            super().render_end()

    def render_advanced(self):
        imgui.push_item_width(40)
        imgui.text("Replace NaN by:")
        _c, self.replace_nan_by = imgui.input_float("##nan", self.replace_nan_by, format = "%.0f")
        self.any_change = self.any_change or _c
        imgui.text("Replace inf by:")
        _c, self.replace_inf_by = imgui.input_float("##inf", self.replace_inf_by, format = "%.0f")
        self.any_change = self.any_change or _c
        imgui.same_line()
        _c, self.replace_inf = imgui.checkbox("replace inf", self.replace_inf)
        self.any_change = self.any_change or _c
        imgui.pop_item_width()

    def get_image_impl(self, idx):
        datasource = self.dataset_in.get_incoming_node()
        if datasource:
            img_in = datasource.get_image(idx)
            pxd = img_in.load()

            if self.operation == 0:
                pxd = pxd ** self.power
            elif self.operation == 1:
                pxd = np.log(pxd)
            elif self.operation == 2:
                pxd = - pxd
            elif self.operation == 3:
                if self.FRAME_REQUESTED_BY_IMAGE_VIEWER:
                    self.last_image_min = np.amin(pxd)
                    self.last_image_max = np.amax(pxd)
                if self.threshold_keep_high:
                    pxd[pxd < self.threshold] = self.threshold_fill_value
                else:
                    pxd[pxd >= self.threshold] = self.threshold_fill_value
            elif self.operation == 4:
                pxd += self.constant
            elif self.operation == 5:
                pxd *= self.factor

            # replace nan and inf
            _inf_val = self.replace_inf_by if self.replace_inf else None
            np.nan_to_num(pxd, copy=False, nan=self.replace_nan_by, posinf = _inf_val, neginf = _inf_val)
            img_out = img_in.clone()
            img_out.data = pxd
            return img_out


from scNodes.core.ceplugin import *
from copy import copy, deepcopy
from pystackreg import StackReg
import cv2
from scipy.ndimage import gaussian_filter

def create():
    return GrayscaleRegPlugin()


class GrayscaleRegPlugin(CEPlugin):
    title = "Register Grayscale"
    description = "Register images of equal or different sizes based on image intensity.\n" \
                  "Useful to register high-magnification TEM images to low-magnification\n" \
                  "images of the same region (e.g. to register Exposure/Search frames onto\n" \
                  "Overview images.) Requires manual picking of the approximate location \n" \
                  "of the child image (the image to be transformed) within parent image." \

    REGMODES = [StackReg.TRANSLATION, StackReg.RIGID_BODY, StackReg.SCALED_ROTATION]
    REGMODES_STR = ["T", "T+R", "T+R+S"]


    def __init__(self):
        self.parent_frame = None
        self.child_frame = None
        self.FLAG_SHOW_LOCATION_PICKER = True
        self.regmode = 0
        self.bin = False
        self.binfac = 2
        self.smooth = False
        self.smoothfac = 2.0
        self.input_pos = True


    def render(self):
        _c, self.parent_frame = self.widget_select_frame_no_rgb("Parent frame:", self.parent_frame)
        _c, self.child_frame = self.widget_select_frame_no_rgb("Child frame:", self.child_frame)

        _cw = imgui.get_content_region_available_width()
        imgui.push_item_width(_cw-120)
        _c, self.regmode = imgui.combo("Transform type", self.regmode, GrayscaleRegPlugin.REGMODES_STR)
        imgui.pop_item_width()
        self.tooltip("T = Translation, R = Rotation, S = Scaling.\n"
                     "Note: when allowing Scaling, the resulting image pixel size may differ from the\n"
                     "value found in the original image's header file or what was input by the user.\n"
                     "Mode T or T+R should be good enough - allowing S tends to worsens results.\n\n"
                     "The selected mode is only applied in the image registration step. In the prior\n"
                     "filtering and rough alignment steps, the input images' (differing) pixel\n"
                     "sizes are always taken in to account.")

        # binning and smoothing
        _c, self.bin = imgui.checkbox("Bin", self.bin)
        if self.bin:
            imgui.same_line(position=110)
            imgui.set_next_item_width(30)
            _c, self.binfac = imgui.input_int("factor##b", self.binfac, 0.0, 0.0)
            self.binfac = max([1, self.binfac])
            self.tooltip("Bin images prior to registration. Factor is the binning factor.")
        _c, self.smooth = imgui.checkbox("Smooth", self.smooth)
        if self.smooth:
            imgui.same_line(position=110)
            imgui.set_next_item_width(30)
            _c, self.smoothfac = imgui.input_float("factor##s", self.smoothfac, 0.0, 0.0, format="%.1f")
            self.smoothfac = max([0.1, self.smoothfac])
            self.tooltip("Apply a Gaussian blur to the images prior to registration.\n"
                         "Factor is the stdev (in units of the child image's pixel size) of the kernel")

        _c, self.input_pos = imgui.checkbox("Location hint:", self.input_pos)
        self.FLAG_SHOW_LOCATION_PICKER = self.input_pos
        if self.input_pos:
            self.widget_selected_position()
        if _c and self.input_pos:
            new_pos = self.child_frame.transform.translation
            cfg.correlation_editor.set_location_indicator_gizmo_pos(new_pos[0], new_pos[1])

        imgui.spacing()
        if self.widget_centred_button("Align!"):
            self.align_frames()


    def align_frames(self):
        try:
            import matplotlib.pyplot as plt
            # Filter, resize, and crop images
            p = self.filter_image(self.parent_frame.data)
            c = self.filter_image(self.child_frame.data)
            npix_p = self.parent_frame.pixel_size
            npix_c = self.child_frame.pixel_size
            _size = np.asarray(np.shape(c)) * npix_c / npix_p
            _size = _size.astype(np.int16) // 2 * 2
            _child = cv2.resize(c, dsize=(_size[1], _size[0]), interpolation=1)  # interpolation=1 == cv2.INTER_LINEAR
            ## GET PIXEL COORDINATES OF WORLD LOCATION CURSOR
            pixel_offset = [0, 0]
            if self.input_pos:
                pixel_coordinate, pixel_offset = self.parent_frame.world_to_pixel_coordinate(self.selected_position)
                _parent = self.crop_around_coordinate(p, _size, pixel_coordinate)
            else:
                _parent = self.crop_center(p, _size)

            if self.bin:
                width, height = _child.shape
                _child = _child[:self.binfac * (width // self.binfac), :self.binfac * (height // self.binfac)]
                _parent = _parent[:self.binfac * (width // self.binfac), :self.binfac * (height // self.binfac)]
                _child = _child.reshape((width // self.binfac, self.binfac, height // self.binfac, self.binfac)).mean(3).mean(1)
                _parent = _parent.reshape((width // self.binfac, self.binfac, height // self.binfac, self.binfac)).mean(3).mean(1)
            if self.smooth:
                _child = gaussian_filter(_child, self.smoothfac)
                _parent = gaussian_filter(_parent, self.smoothfac)

            # Find transformation matrix that matches the frames
            sr = StackReg(GrayscaleRegPlugin.REGMODES[self.regmode])
            tmat = sr.register(_parent, _child)
            T, R, S = self.decompose_transform_matrix(tmat)

            # Apply transform to child
            self.child_frame.parent_to(self.parent_frame)
            self.child_frame.transform = deepcopy(self.parent_frame.transform)
            dx = T[0] * npix_c + pixel_offset[0] * npix_p
            dy = T[1] * npix_c + pixel_offset[1] * npix_p
            self.child_frame.transform.translation[0] += dx
            self.child_frame.transform.translation[1] += dy
            self.child_frame.transform.rotation -= R
            self.child_frame.transform.scale *= S
            self.child_frame.pivot_point[0] = self.child_frame.transform.translation[0]
            self.child_frame.pivot_point[1] = self.child_frame.transform.translation[1]

        except Exception as e:
            cfg.set_error(e, "Error aligning frames in Register Grayscale tool.")


    @staticmethod
    def crop_center(img, size):
        w, h = img.shape
        dw = (w - size[0]) // 2
        dh = (h - size[1]) // 2
        return img[dw:dw + size[0], dh:dh + size[1]]


    @staticmethod
    def crop_around_coordinate(img, size, coordinate):
        W = size[0]
        H = size[1]
        w, h = img.shape
        if W > w or H > h:
            raise Exception("ROI can't be larger than the original image.")
        X = coordinate[1]
        Y = coordinate[0]
        X = max([W // 2, min([X, w - W // 2])])
        Y = max([H // 2, min([Y, h - H // 2])])
        return img[X - W // 2:X + W // 2, Y - H // 2:Y + H // 2]


    @staticmethod
    def filter_image(array):
        data = copy(array)
        mu = np.mean(data)
        std = np.std(data)
        mask = data > (mu + std * 10.0)
        data[mask] = mu
        data = data - np.amin(data)
        data = data / np.amax(data)
        return data


    @staticmethod
    def decompose_transform_matrix(tmat):
        vec = np.matrix(tmat[0:2, 0:2]) * np.matrix([[1], [0]])
        scaling = float(np.sqrt(vec[0] ** 2 + vec[1] ** 2))
        rotation = float(np.arctan2(vec[1], vec[0]) * 360.0 / 2.0 / np.pi)
        translation = np.array([tmat[0, 2], tmat[1, 2]])
        return translation, rotation, scaling



from OpenGL.GL import *
from OpenGL.GLUT import *
from itertools import count
import numpy as np
import ctypes

"""All classes here are fairly straightforward implementations of standard OpenGL objects. Some notes:
in Texture/Framebuffer constructor there is a format parameter, which can be any of 'rgbu16', 'ru16', or 'rgba32f'.
For vertex and index buffers, the input vertices/indices data has to be a default python list [x1, y1, z1, x2, y2, ...] 
or [face11, face12, face13, face21, face22, etc.] (for vertex resp. index buffers)
All classes have bind() and unbind()* functions. (* no unbind for Texture). Texture.bind(slot) takes an argument, 
int:slot, to allow binding to any texture slot. When doing so, remember to set glActiveTexture(GL_TEXTURE0) later on 
whenever necessary.
"""

class Texture:
    idGenerator = count()

    IF_F_T = dict() # internalformat, format, type
    IF_F_T[None] = [GL_R16UI, GL_RED_INTEGER, GL_UNSIGNED_SHORT]
    IF_F_T["rgba32f"] = [GL_RGBA32F, GL_RGBA, GL_FLOAT]
    IF_F_T["rgb32f"] = [GL_RGB32F, GL_RGB, GL_FLOAT]
    IF_F_T["r32f"] = [GL_R32F, GL_RED, GL_FLOAT]
    IF_F_T["rgbu16"] = [GL_RGB16UI, GL_RGB_INTEGER, GL_UNSIGNED_SHORT]
    IF_F_T["rgb8u"] = [GL_RGB8UI, GL_RGB_INTEGER, GL_UNSIGNED_INT]  # should be GL_BYTE but not touching it for now
    IF_F_T["rgba8u"] = [GL_RGBA8UI, GL_RGBA_INTEGER, GL_UNSIGNED_INT]  # should be GL_BYTE but not touching it for now
    IF_F_T["rgba16u"] = [GL_RGBA16UI, GL_RGBA_INTEGER, GL_UNSIGNED_SHORT]
    IF_F_T["rgba32u"] = [GL_RGBA32UI, GL_RGBA_INTEGER, GL_UNSIGNED_INT]

    def __init__(self, format=None):
        if format:
            self.id = next(Texture.idGenerator)
            self.renderer_id = glGenTextures(1)
            self.width = 0
            self.height = 0
            if_f_t = Texture.IF_F_T[format]
            self.internalformat = if_f_t[0]
            self.format = if_f_t[1]
            self.type = if_f_t[2]

            glBindTexture(GL_TEXTURE_2D, self.renderer_id)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)

    def bind(self, slot=0):
        glActiveTexture(GL_TEXTURE0 + int(slot))
        glBindTexture(GL_TEXTURE_2D, self.renderer_id)

    def update(self, pixeldata, width=None, height=None):
        self.bind()
        if width:
            self.width = width
        if height:
            self.height = height
        else:
            self.width = np.shape(pixeldata)[1]
            self.height = np.shape(pixeldata)[0]

        imgdata = None
        if not pixeldata is None:
            imgdata = pixeldata.flatten()

        glTexImage2D(GL_TEXTURE_2D, 0, self.internalformat, self.width, self.height, 0, self.format, self.type, imgdata)

    def set_linear_interpolation(self):
        glBindTexture(GL_TEXTURE_2D, self.renderer_id)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)

    def set_no_interpolation(self):
        glBindTexture(GL_TEXTURE_2D, self.renderer_id)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)

class Shader:
    """Uniforms can only be uploaded when Shaders is first manually bound by user."""
    def __init__(self, sourcecode = None):
        if sourcecode:
            self.shaderProgram = glCreateProgram()
            self.compile(sourcecode)

    def compile(self, sourcecode):
        shaderStages = dict()
        shaderObjects = list()
        currentStage = None
        # Parse source into individual shader stages' source code
        with open(sourcecode, 'r') as source:
            for line in source:
                if "#vertex" in line:
                    currentStage = GL_VERTEX_SHADER
                    shaderStages[currentStage] = ""
                elif "#fragment" in line:
                    currentStage = GL_FRAGMENT_SHADER
                    shaderStages[currentStage] = ""
                elif "#geometry" in line:
                    currentStage = GL_GEOMETRY_SHADER
                    shaderStages[currentStage] = ""
                elif "#compute" in line:
                    currentStage = GL_COMPUTE_SHADER
                    shaderStages[currentStage] = ""
                else:
                    shaderStages[currentStage] += line
        # Compile stages
        for key in shaderStages:
            shaderObjects.append(glCreateShader(key))
            glShaderSource(shaderObjects[-1], shaderStages[key])
            glCompileShader(shaderObjects[-1])
            status = glGetShaderiv(shaderObjects[-1], GL_COMPILE_STATUS)
            if status == GL_FALSE:
                if key == GL_VERTEX_SHADER:
                    strShaderType = "vertex"
                elif key == GL_FRAGMENT_SHADER:
                    strShaderType = "fragment"
                elif key == GL_GEOMETRY_SHADER:
                    strShaderType = "geometry"
                elif key == GL_COMPUTE_SHADER:
                    strShaderType = "compute"
                raise RuntimeError("Shaders compilation failure for type "+strShaderType+":\n" + glGetShaderInfoLog(shaderObjects[-1]).decode('utf-8'))
            glAttachShader(self.shaderProgram, shaderObjects[-1])
        glLinkProgram(self.shaderProgram)
        status = glGetProgramiv(self.shaderProgram, GL_LINK_STATUS)
        if status == GL_FALSE:
            raise RuntimeError("Shaders link failure:\n"+glGetProgramInfoLog(self.shaderProgram).decode('utf-8'))
        for shader in shaderObjects:
            glDetachShader(self.shaderProgram, shader)
            glDeleteShader(shader)

    def bind(self):
        glUseProgram(self.shaderProgram)

    def unbind(self):
        glUseProgram(0)

    def uniform1f(self, uniformName, uniformFloatValue):
        uniformLocation = glGetUniformLocation(self.shaderProgram, uniformName)
        glUniform1f(uniformLocation, uniformFloatValue)

    def uniform1i(self, uniformName, uniformIntValue):
        uniformLocation = glGetUniformLocation(self.shaderProgram, uniformName)
        glUniform1i(uniformLocation, uniformIntValue)

    def uniform2f(self, uniformName, uniformFloat2Value):
        uniformLocation = glGetUniformLocation(self.shaderProgram, uniformName)
        glUniform2f(uniformLocation, uniformFloat2Value[0], uniformFloat2Value[1])

    def uniform3f(self, uniformName, uniformFloat3Value):
        uniformLocation = glGetUniformLocation(self.shaderProgram, uniformName)
        glUniform3f(uniformLocation, uniformFloat3Value[0], uniformFloat3Value[1], uniformFloat3Value[2])

    def uniformmat4(self, uniformName, uniformMat4):
        uniformLocation = glGetUniformLocation(self.shaderProgram, uniformName)
        glUniformMatrix4fv(uniformLocation, 1, GL_TRUE, uniformMat4)

class VertexBuffer:
    """Not that vertices must be a default 1d python list. In __init__ it is cast into the required shape."""
    def __init__(self, vertex_data = None):
        self.vertexBufferObject = glGenBuffers(1)
        self.location = 0
        self.stride = 1
        if not vertex_data is None:
            self.update(vertex_data)

    def update(self, vertex_data):
        _data = np.asarray([[vertex_data]], dtype=np.float32)
        glBindBuffer(GL_ARRAY_BUFFER, self.vertexBufferObject)
        glBufferData(GL_ARRAY_BUFFER, _data, GL_STATIC_DRAW)
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def bind(self):
        glBindBuffer(GL_ARRAY_BUFFER, self.vertexBufferObject)

    def unbind(self):
        glBindBuffer(GL_ARRAY_BUFFER, 0)

    def set_location_and_stride(self, location, stride):
        self.location = location
        self.stride = stride
        glEnableVertexAttribArray(location)
        self.bind()
        glVertexAttribPointer(location, stride, GL_FLOAT, GL_FALSE, stride * 4, None)
        self.unbind()

    def set_divisor_to_per_instance(self):
        glVertexAttribDivisor(self.location, 1)


class IndexBuffer:
    """Note that indices must be a default python list. It is turned in to a np.array along the 2nd dimension with type np.uint16 before sending to GPU"""
    def __init__(self, indices):
        self.indexBufferObject = glGenBuffers(1)
        self.indices = np.asarray([indices], dtype = np.uint16)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.indexBufferObject)
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices, GL_STATIC_DRAW)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

    def bind(self):
        glBindBuffer(self.indexBufferObject)

    def unbind(self):
        glBindBuffer(0)

    def getCount(self):
        return self.indices.size


class VertexArray:
    def __init__(self, vertexBuffer = None, indexBuffer = None, attribute_format="xyzuv"):
        self.vertexBuffer = None
        self.indexBuffer = None
        self.initialized = False
        self.attribute_format = attribute_format
        self.vertexArrayObject = None
        if vertexBuffer:
            self.update(vertexBuffer, indexBuffer)

    def init(self):
        if self.initialized:
            return
        else:
            self.vertexArrayObject = glGenVertexArrays(1)
            self.initialized = True

    def update(self, vertexBuffer, indexBuffer):
        if not self.initialized:
            self.init()
        self.vertexBuffer = vertexBuffer
        self.indexBuffer = indexBuffer
        glBindVertexArray(self.vertexArrayObject)
        glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer.vertexBufferObject)
        if self.attribute_format == "xyzuv":
            glEnableVertexAttribArray(0)
            glEnableVertexAttribArray(1)
            glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 20, ctypes.cast(0, ctypes.c_void_p))
            glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 20, ctypes.cast(12, ctypes.c_void_p))
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer.indexBufferObject)
        elif self.attribute_format == "xy":
            glEnableVertexAttribArray(0)
            glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 8, ctypes.cast(0, ctypes.c_void_p))
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer.indexBufferObject)
        elif self.attribute_format == "xyuv":
            glEnableVertexAttribArray(0)
            glEnableVertexAttribArray(1)
            glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 16, ctypes.cast(0, ctypes.c_void_p))
            glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 16, ctypes.cast(8, ctypes.c_void_p))
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBuffer.indexBufferObject)
        glBindVertexArray(0)

    def bind(self):
        glBindVertexArray(self.vertexArrayObject)

    def unbind(self):
        glBindVertexArray(0)


class FrameBuffer:
    """For now, FrameBuffers have one texture (COLOR and DEPTH) only."""
    def __init__(self, width=None, height=None, texture_format="rgba32f"):
        """
        :param width: int
        :param height: int
        :param texture_format: format of the render texture. One of the texture formats allowed in Texture() (see above)
        """
        if width:
            # Set up internal parameters
            self.width = width
            self.height = height
            # Set up texture
            self.texture_format = texture_format
            self.texture = Texture(format=texture_format)
            self.texture.bind()
            self.texture.update(None, self.width, self.height)
            glBindTexture(GL_TEXTURE_2D, 0)
            # Set up depth render buffer
            self.depthRenderbuffer = glGenRenderbuffers(1)
            glBindRenderbuffer(GL_RENDERBUFFER, self.depthRenderbuffer)
            glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, self.width, self.height)
            glBindRenderbuffer(GL_RENDERBUFFER, 0)
            # Set up frame buffer
            self.framebufferObject = glGenFramebuffers(1)
            glBindFramebuffer(GL_FRAMEBUFFER, self.framebufferObject)
            glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, self.texture.renderer_id, 0)
            glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, self.depthRenderbuffer)

            if glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE:
                raise RuntimeError("Framebuffer binding failed, GPU might not support this configuration.")
            glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def clear(self, color):
        glBindFramebuffer(GL_FRAMEBUFFER, self.framebufferObject)
        glClearColor(color[0], color[1], color[2], 1.0 if len(color) == 3 else color[3])
        glClearDepth(1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glBindFramebuffer(GL_FRAMEBUFFER, 0)

    def bind(self, auto_set_viewport = True):
        glBindFramebuffer(GL_FRAMEBUFFER, self.framebufferObject)
        if auto_set_viewport:
            glViewport(0, 0, self.width, self.height)

    def unbind(self, viewport = None):
        """When the FBO was bound, the viewport may have been changed. Remember to change it back to the right size! Forgetting to do so does not cause errors, but subsequent viewports may appear clipped. Viewport:tuple(int, int, int, int)"""
        glBindFramebuffer(GL_FRAMEBUFFER, 0)
        if viewport:
            glViewport(*viewport)

    def __str__(self):
        return f"FrameBuffer object with colour and depth texture. Colour texture format is {self.texture_format}, size is {self.width} x {self.height}"


from scNodes.core.ceplugin import *


def create():
    return CustomPlugin()


class CustomPlugin(CEPlugin):
    title = "Custom plugin template"
    description = "The description to be shown when the '?' icon is hovered goes here"
    enabled = False  # default is True. Set to False in order not to load the plugin upon booting the software.
    FLAG_SHOW_LOCATION_PICKER = True  # default is False. Set to True to enable a location marker for this plugin (as in the Register Grayscale node)

    def __init__(self):
        ## variables for settings can be initialized here.
        pass

    def render(self):
        # body of the node goes here.
        # see ceplugin.py for a list of predefined widgets that may be useful.
        # these can be called as in the following examples:

        self.widget_show_active_frame_title() # text field showing the title of the currently selected frame

        _selection_changed, frame = self.widget_select_frame_any("Select frame", 0)  # drop-down menu with all frames currently in the scene available for selection. _selection_changed is a bool flagging whether the selected frame was changed
        _selection_changed, frame = self.widget_select_frame_no_rgb("Select frame", 0)  # drop-down menu with all grayscale frames currently in the scene available for selection
        _selection_changed, frame = self.widget_select_frame_rgb("Select frame", 0)  # drop-down menu with all rgb frames currently in the scene available for selection

        self.widget_selected_position()  # shows the current coordinates of the location picker

        if self.widget_centred_button("Click me"):  # adds a button that is horizontally centred in the window
            self.process_frame(frame)
        self.tooltip("This text shows when the button is hovered.")

    @staticmethod
    def process_frame(frame):
        pass  # this function is called when the user presses the 'Click me' button that was defined above. Functionality of the plugin may go here (or simply within the if clause in the render function)



import importlib.util
import shutil
import imgui
from imgui.integrations.glfw import GlfwRenderer
import glfw
import sys
import tkinter as tk
from tkinter import filedialog
from scNodes.core.node import *
import os
import pyperclip
import dill as pickle
from scNodes.core.opengl_classes import Texture
tkroot = tk.Tk()
tkroot.withdraw()


class NodeEditor:
    TOOLTIP_APPEAR_DELAY = 1.0  # seconds
    TOOLTIP_HOVERED_TIMER = 0.0
    TOOLTIP_HOVERED_START_TIME = 0.0

    NODE_FACTORY = dict()
    NODE_GROUPS = dict()

    def __init__(self, window, imgui_context, imgui_impl):
        self.window = window
        self.window.clear_color = cfg.COLOUR_WINDOW_BACKGROUND
        self.window.make_current()

        self.imgui_context = imgui_context
        self.imgui_implementation = imgui_impl


        # Context menu
        self.context_menu_position = [0, 0]
        self.context_menu_open = False
        self.context_menu_can_close = False

        NodeEditor.init_node_factory()

        if True:
            self.boot_img_texture = Texture(format="rgba32f")
            pxd_boot_img_texture = np.asarray(Image.open(cfg.root+"icons/scnodes_boot_img.png")).astype(np.float32) / 255.0
            self.boot_img_texture.update(pxd_boot_img_texture)
            self.boot_img_height, self.boot_img_width = pxd_boot_img_texture.shape[0:2]
            self.boot_img_texture.set_linear_interpolation()
            self.show_boot_img = True

    def get_font_atlas_ptr(self):
        return self.imgui_implementation.io.fonts

    def on_update(self):
        if cfg.node_editor_relink:
            NodeEditor.relink_after_load()
            cfg.node_editor_relink = False
        if cfg.next_active_node is not None:
            cfg.active_node = cfg.next_active_node
            cfg.next_active_node = None

        imgui.set_current_context(self.imgui_context)
        self.window.make_current()
        self.window.set_full_viewport()
        if self.window.focused:
            self.imgui_implementation.process_inputs()
        self.window.on_update()
        if self.window.window_size_changed:
            cfg.window_width = self.window.width
            cfg.window_height = self.window.height
        if not imgui.get_io().want_capture_keyboard and imgui.is_key_pressed(glfw.KEY_TAB):
            cfg.active_editor = 1
        if not self.window.get_key(glfw.KEY_ESCAPE):
            for node in cfg.nodes:
                node.clear_flags()
                node.on_update()
        else:
            pass
        if cfg.focused_node is not None:
            cfg.focused_node.any_change = cfg.any_change
        cfg.any_change = False
        cfg.connector_delete_requested = False
        cfg.active_connector_hover_pos = self.window.cursor_pos
        if cfg.active_connector is not None:
            if not self.window.get_mouse_button(glfw.MOUSE_BUTTON_LEFT):
                cfg.connector_released = True
                cfg.active_connector_hover_pos = None
                cfg.active_connector = None
            else:
                cfg.connector_released = False
        if self.window.get_mouse_event(glfw.MOUSE_BUTTON_RIGHT, glfw.PRESS, pop_event=False):
            cfg.connector_delete_requested = True

        cfg.node_move_requested = [0, 0]
        cfg.camera_move_requested = [0, 0]
        imgui_want_mouse = imgui.get_io().want_capture_mouse

        if self.window.get_mouse_button(glfw.MOUSE_BUTTON_LEFT) and imgui_want_mouse:
            cfg.node_move_requested = self.window.cursor_delta
        elif self.window.get_mouse_button(glfw.MOUSE_BUTTON_MIDDLE) and not imgui_want_mouse:
            cfg.camera_move_requested = self.window.cursor_delta

        imgui.get_io().display_size = self.window.width, self.window.height
        imgui.new_frame()
        self._gui_main()
        imgui.render()
        self.imgui_implementation.render(imgui.get_draw_data())

    def end_frame(self):
        self.window.end_frame()

    def _gui_main(self):
        # Render nodes  - render active_connector_parent_node first, to enable all other connectors' drop targets.
        source_node_id = -1
        if cfg.active_connector_parent_node is not None:
            cfg.active_connector_parent_node.render()
            source_node_id = cfg.active_connector_parent_node.id
        for node in cfg.nodes:
            if node.id is not source_node_id:
                node.render()

        ## Context menu
        if not self.context_menu_open:
            if self.window.get_mouse_event(glfw.MOUSE_BUTTON_RIGHT, glfw.PRESS, 0) and not imgui.get_io().want_capture_mouse:
                self.context_menu_position = self.window.cursor_pos
                self.context_menu_open = True
                self.context_menu_can_close = False
        else:
            self._context_menu()

        NodeEditor._menu_bar()
        self._warning_window()
        self._boot_img()

    def _boot_img(self):
        if self.show_boot_img:
            imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND, *cfg.COLOUR_WINDOW_BACKGROUND[0:3], 0.0)
            imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_ACTIVE, *cfg.COLOUR_WINDOW_BACKGROUND[0:3], 0.0)
            imgui.push_style_color(imgui.COLOR_TEXT, *(0.0, 0.0, 0.0, 1.0))

            _w = self.boot_img_width * 0.5
            _h = self.boot_img_height * 0.5
            imgui.set_next_window_position((cfg.window_width - _w) / 2.0, (cfg.window_height - _h) / 2.0 - 25)
            self.show_boot_img = imgui.begin("##bootwindow", True, imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_ALWAYS_AUTO_RESIZE | imgui.WINDOW_NO_BACKGROUND)[1]
            imgui.image(self.boot_img_texture.renderer_id, _w, _h)
            imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *cfg.COLOUR_WINDOW_BACKGROUND)
            if imgui.begin_popup_context_window():
                imgui.text(f"version {cfg.version}\nsource: github.com/bionanopatterning/scNodes")
                imgui.end_popup()
            imgui.pop_style_color(1)
            imgui.end()
            imgui.pop_style_color(3)

    def _warning_window(self):
        def ww_context_menu():
            imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *cfg.COLOUR_MENU_WINDOW_BACKGROUND)
            if imgui.begin_popup_context_window():
                raise_error, _ = imgui.menu_item("Raise error (debug)")
                if raise_error:
                    raise cfg.error_obj
                copy_error, _ = imgui.menu_item("Copy to clipboard")
                if copy_error:
                    pyperclip.copy(cfg.error_msg)
                imgui.end_popup()
            imgui.pop_style_color(1)
        ## Error message
        if cfg.error_msg is not None:
            if cfg.error_new:
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND, *cfg.COLOUR_ERROR_WINDOW_HEADER_NEW)
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_ACTIVE, *cfg.COLOUR_ERROR_WINDOW_HEADER_NEW)
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_COLLAPSED, *cfg.COLOUR_ERROR_WINDOW_HEADER_NEW)
            else:
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND, *cfg.COLOUR_ERROR_WINDOW_HEADER)
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_ACTIVE, *cfg.COLOUR_ERROR_WINDOW_HEADER)
                imgui.push_style_color(imgui.COLOR_TITLE_BACKGROUND_COLLAPSED, *cfg.COLOUR_ERROR_WINDOW_HEADER)
            imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *cfg.COLOUR_ERROR_WINDOW_BACKGROUND)
            imgui.push_style_color(imgui.COLOR_TEXT, *cfg.COLOUR_ERROR_WINDOW_TEXT)
            imgui.push_style_var(imgui.STYLE_WINDOW_ROUNDING, 3.0)
            imgui.set_next_window_size(self.window.width, cfg.ERROR_WINDOW_HEIGHT)
            imgui.set_next_window_position(0, self.window.height - cfg.ERROR_WINDOW_HEIGHT)
            _, stay_open = imgui.begin("Notifications", True, imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_COLLAPSE)
            if not cfg.error_logged:
                cfg.write_to_log(cfg.error_msg)
                cfg.error_logged = True
            imgui.text(cfg.error_msg)
            if imgui.is_window_focused() and self.window.get_mouse_event(glfw.MOUSE_BUTTON_LEFT, glfw.PRESS):
                cfg.error_new = False
            ww_context_menu()
            imgui.end()
            if not stay_open:
                cfg.error_msg = None
                cfg.error_new = True
            imgui.pop_style_color(5)
            imgui.pop_style_var(1)

    def _context_menu(self):
        imgui.set_next_window_position(self.context_menu_position[0] - 3, self.context_menu_position[1] - 3)
        imgui.push_style_color(imgui.COLOR_WINDOW_BACKGROUND, *cfg.COLOUR_MENU_WINDOW_BACKGROUND)
        imgui.push_style_color(imgui.COLOR_TEXT, *cfg.COLOUR_CM_WINDOW_TEXT)
        imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *cfg.COLOUR_MENU_WINDOW_BACKGROUND)
        imgui.begin("##necontextmenu", flags=imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_ALWAYS_AUTO_RESIZE)
        # Close context menu when it is not hovered.
        context_menu_hovered = imgui.is_window_hovered(flags=imgui.HOVERED_ALLOW_WHEN_BLOCKED_BY_POPUP | imgui.HOVERED_CHILD_WINDOWS)
        if context_menu_hovered:
            self.context_menu_can_close = True
        if not context_menu_hovered and self.context_menu_can_close:
            self.context_menu_open = False

        # Context menu contents
        new_node = None
        for key in NodeEditor.NODE_GROUPS:
            if imgui.begin_menu(key):
                for nodetitle in NodeEditor.NODE_GROUPS[key]:
                    item_selected, _ = imgui.menu_item(nodetitle)
                    if item_selected:
                        new_node = NodeEditor.NODE_FACTORY[nodetitle]()
                imgui.end_menu()
        if new_node:
            try:
                new_node.position = self.context_menu_position
                cfg.set_active_node(new_node)
                self.context_menu_open = False
            except Exception as e:
                cfg.set_error(e, "Error upon requesting new node (probably not implemented yet.)\nError:"+str(e))

        if cfg.focused_node is not None:
            clear_active_node, _ = imgui.menu_item("Release focused node")
            if clear_active_node:
                cfg.set_active_node(None, True)
        # End
        imgui.pop_style_color(3)
        imgui.end()

    @staticmethod
    def _menu_bar():
        imgui.push_style_color(imgui.COLOR_MENUBAR_BACKGROUND, *cfg.COLOUR_MAIN_MENU_BAR)
        imgui.push_style_color(imgui.COLOR_TEXT, *cfg.COLOUR_MAIN_MENU_BAR_TEXT)
        imgui.push_style_color(imgui.COLOR_HEADER_HOVERED, *cfg.COLOUR_MAIN_MENU_BAR_HILIGHT)
        imgui.push_style_color(imgui.COLOR_HEADER_ACTIVE, *cfg.COLOUR_MAIN_MENU_BAR_HILIGHT)
        imgui.push_style_color(imgui.COLOR_HEADER, *cfg.COLOUR_MAIN_MENU_BAR_HILIGHT)
        imgui.push_style_color(imgui.COLOR_POPUP_BACKGROUND, *cfg.COLOUR_MENU_WINDOW_BACKGROUND)
        ## Save node setup.
        if imgui.core.begin_main_menu_bar():
            if imgui.begin_menu('File'):
                if imgui.menu_item("Save project")[0]:
                    try:
                        filename = filedialog.asksaveasfilename(filetypes=[("srNodes project", ".srnp")])
                        if filename != '':
                            cfg.save_project(filename)
                    except Exception as e:
                        cfg.set_error(e, f"Error saving project\n")
                if imgui.menu_item("Load project")[0]:
                    try:
                        filename = filedialog.askopenfilename(filetypes=[("srNodes project", ".srnp")])
                        if filename != '':
                            cfg.load_project(filename)
                    except Exception as e:
                        cfg.set_error(e, f"Error loading project - are you sure you selected a '.srnp' file?\n")
                load_setup, _ = imgui.menu_item("Load node setup")
                if load_setup:
                    try:
                        filename = filedialog.askopenfilename(filetypes=[("srNodes setup", ".srn")])
                        if filename != '':
                            with open(filename, 'rb') as pickle_file:
                                imported_nodes = pickle.load(pickle_file)
                                cfg.nodes = imported_nodes
                                cfg.node_editor_relink = True
                    except Exception as e:
                        cfg.set_error(e, f"Error loading node setup - are you sure you selected a '.srn' file?\n"+str(e))
                import_setup, _ = imgui.menu_item("Append node setup")
                if import_setup:
                    try:
                        filename = filedialog.askopenfilename(filetypes=[("srNodes setup", ".srn")])
                        if filename != '':
                            with open(filename, 'rb') as pickle_file:
                                imported_nodes = pickle.load(pickle_file)
                                for node in imported_nodes:
                                    cfg.nodes.append(node)
                                NodeEditor.relink_after_load()
                    except Exception as e:
                        cfg.set_error(e, "Error importing node setup - are you sure you selected a '.srn' file?\n"+str(e))
                save_setup, _ = imgui.menu_item("Save current setup")
                if save_setup:
                    filename = filedialog.asksaveasfilename(filetypes=[("srNodes setup", ".srn")])
                    if filename != '':
                        if filename[-4:] == ".srn":
                            filename = filename[:-4]
                        with open(filename+".srn", 'wb') as pickle_file:
                            node_copies = list()
                            for node in cfg.nodes:
                                node.pre_save()
                                node_copies.append(copy.deepcopy(node))
                                node.post_save()
                            pickle.dump(node_copies, pickle_file)
                clear_setup, _ = imgui.menu_item("Clear current setup")
                if clear_setup:
                    for i in range(len(cfg.nodes)):
                        cfg.nodes[0].delete()
                    cfg.nodes = list()
                imgui.end_menu()
            if imgui.begin_menu('Settings'):
                install_node, _ = imgui.menu_item("Install a node")
                if install_node:
                    NodeEditor.install_node()
                if imgui.begin_menu('Profiling'):
                    _c, cfg.profiling = imgui.checkbox("Track node processing times", cfg.profiling)
                    NodeEditor.tooltip("Keep track of the time that every node in the pipeline takes to process and output a frame.\n")
                    clear_times, _ = imgui.menu_item("   Reset timers")
                    if clear_times:
                        for node in cfg.nodes:
                            node.profiler_time = 0.0
                            node.profiler_count = 0
                    imgui.end_menu()
                if imgui.begin_menu('Parallel processing'):
                    imgui.set_next_item_width(26)
                    _c, cfg.n_cpus = imgui.input_int("batch size", cfg.n_cpus, 0, 0)
                    if _c:
                        cfg.batch_size = cfg.n_cpus
                    NodeEditor.tooltip(
                        "Number of frames to process within one parallel processing batch. Values higher than the amount\n"
                        "of CPUs on the PC are allowed and will result in multiple tasks being dispatched to individual\n"
                        "CPUs per batch. This can increase processing speed, but reduces GUI responsiveness. For optimal\n"
                        "efficiency, set the batch size to an integer multiple of the amount of CPUs on the machine. \n"
                        f"This PC has: {cfg.n_cpus_max} CPUs.\n"
                        f"Set to '-1' to force use of all CPUs.")
                    imgui.end_menu()
                imgui.end_menu()
            if imgui.begin_menu('Editor'):
                select_node_editor, _ = imgui.menu_item("Node Editor", None, selected=True)
                select_correlation_editor, _ = imgui.menu_item("Correlation", None, selected=False)
                if select_correlation_editor:
                    cfg.active_editor = 1
                imgui.end_menu()
            imgui.end_main_menu_bar()
        imgui.pop_style_color(6)

    @staticmethod
    def install_node():
        try:
            filename = filedialog.askopenfilename(filetypes=[("Python file", ".py")])
            if filename != '':
                node_dir = __file__[:__file__.rfind("\\")]+"/nodes/"
                node_dir = node_dir.replace('\\', '/')
                node_name = filename[filename.rfind("/")+1:]
                shutil.copyfile(filename, node_dir+node_name)
            cfg.set_error(Exception(), "Node installed! Restarting the software is required for it to become available.\n\n\n")
        except Exception as e:
            cfg.set_error(e, "Error upon installing node. Are you sure you selected the right file?")

    @staticmethod
    def init_node_factory():
        nodeimpls = list()

        class NodeImpl:
            def __init__(self, node_create_fn):
                self.create_fn = node_create_fn
                self.node_obj = node_create_fn()
                self.title = self.node_obj.title
                self.group = self.node_obj.group
                self.id = self.node_obj.sortid
                self.enabled = self.node_obj.enabled

            def __del__(self):
                self.node_obj.delete()

        node_source_files = glob.glob(cfg.root+"nodes/*.py")  # load all .py files in the /nodes folder
        i = 0
        for nodesrc in node_source_files:  # for every file, dynamically load module and save the module's create() function to a dict, keyed by name of node.
            i += 1
            if "__init__.py" in nodesrc:
                continue

            module_name = nodesrc[nodesrc.rfind("\\")+1:-3]
            try:
                # get the module spec and import the module
                mod = importlib.import_module("scNodes.nodes."+module_name)
                impl = NodeImpl(mod.create)
                if not impl.enabled:
                    continue
                nodeimpls.append(impl)

            except Exception as e:
                cfg.nodes = list()
                cfg.set_error(e, f"No well-defined Node type found in {nodesrc}. See manual for minimal code requirements.")
        node_ids = list()

        for ni in nodeimpls:
            node_ids.append(ni.id)
        sorted_id_indices = np.argsort(node_ids)

        for nid in sorted_id_indices:
            _node = nodeimpls[nid]
            NodeEditor.NODE_FACTORY[_node.title] = _node.create_fn
            if isinstance(_node.group, str):
                if _node.group not in NodeEditor.NODE_GROUPS.keys():
                    NodeEditor.NODE_GROUPS[_node.group] = list()
                NodeEditor.NODE_GROUPS[_node.group].append(_node.title)
            elif isinstance(_node.group, list):
                for group in _node.group:
                    if group not in NodeEditor.NODE_GROUPS.keys():
                        NodeEditor.NODE_GROUPS[group] = list()
                    NodeEditor.NODE_GROUPS[group].append(_node.title)

        NodeEditor.node_group_all = list(NodeEditor.NODE_FACTORY.keys())

    @staticmethod
    def relink_after_load():
        nodes = cfg.nodes
        attributes = list()
        ids_to_link = list()
        for node in nodes:
            for attribute in node.connectable_attributes:
                attributes.append(attribute)
                if attribute.direction == ConnectableAttribute.INPUT:
                    for partner in attribute.linked_attributes:
                        ids_to_link.append((attribute.id, partner.id))
                #attribute.disconnect_all()

        def _find_attribute_by_id(target_id):
            for a in attributes:
                if a.id == target_id:
                    return a

        for pair in ids_to_link:
            ca_a = _find_attribute_by_id(pair[0])
            ca_b = _find_attribute_by_id(pair[1])
            ca_a.connect_attributes(ca_b)

    @staticmethod
    def delete_temporary_files():
        dirs = glob.glob("_srnodes_temp*/")
        for dir in dirs:
            shutil.rmtree(dir)

    @staticmethod
    def tooltip(text):
        if imgui.is_item_hovered():
            if NodeEditor.TOOLTIP_HOVERED_TIMER == 0.0:
                NodeEditor.TOOLTIP_HOVERED_START_TIME = time.time()
                NodeEditor.TOOLTIP_HOVERED_TIMER = 0.001  # add a fake 1 ms to get out of this if clause
            elif NodeEditor.TOOLTIP_HOVERED_TIMER > NodeEditor.TOOLTIP_APPEAR_DELAY:
                imgui.set_tooltip(text)
            else:
                NodeEditor.TOOLTIP_HOVERED_TIMER = time.time() - NodeEditor.TOOLTIP_HOVERED_START_TIME
        if not imgui.is_any_item_hovered():
            NodeEditor.TOOLTIP_HOVERED_TIMER = 0.0




from scNodes.core.node import *


def create():
    return ParticleFilterNode()


class ParticleFilterNode(Node):
    title = "Particle filter"
    group = "PSF-fitting reconstruction"
    colour = (243 / 255, 0 / 255, 80 / 255, 1.0)
    size = 310
    sortid = 1002
    def __init__(self):
        super().__init__()

        self.reconstruction_in = ConnectableAttribute(ConnectableAttribute.TYPE_RECONSTRUCTION, ConnectableAttribute.INPUT, parent=self)
        self.reconstruction_out = ConnectableAttribute(ConnectableAttribute.TYPE_RECONSTRUCTION, ConnectableAttribute.OUTPUT, parent=self)

        self.available_parameters = ["No data available"]
        self.filters = list()

        self.returns_image = False
        self.does_profiling_count = False

    def render(self):
        if super().render_start():
            self.reconstruction_in.render_start()
            self.reconstruction_out.render_start()
            self.reconstruction_in.render_end()
            self.reconstruction_out.render_end()

            imgui.spacing()
            imgui.separator()
            imgui.spacing()

            if self.reconstruction_in.check_connect_event():
                self.get_histogram_parameters()

            i = 0
            imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM, 0.2, 0.2, 0.2, 1.0)
            imgui.push_style_color(imgui.COLOR_PLOT_HISTOGRAM_HOVERED, 0.2, 0.2, 0.2, 1.0)
            for pf in self.filters:
                imgui.push_id(f"pfid{i}")
                if imgui.button("-", 20, 20):
                    self.filters.remove(pf)
                    imgui.pop_id()
                    continue
                imgui.same_line()
                imgui.set_next_item_width(144)
                _c, pf.parameter = imgui.combo("Criterion", pf.parameter, self.available_parameters)
                if _c:
                    pf.set_data(*self.get_histogram_vals(self.available_parameters[pf.parameter]), self.available_parameters[pf.parameter])
                imgui.same_line(spacing = 10)
                _c, pf.invert = imgui.checkbox("NOT", pf.invert)
                Node.tooltip("If NOT is selected, the filter logic is inverted. Default\n"
                             "behaviour is: particle retained if min < parameter < max.\n"
                             "With NOT on, a particle is retained if max < parameter OR\nmin > parameter")

                pf.render()
                imgui.pop_id()
                i += 1
            content_width = imgui.get_window_content_region_width()
            imgui.new_line()
            imgui.same_line(content_width / 2 - 50)
            if imgui.button("Add filter", width = 100, height = 20):
                new_filter = ParticleFilterNode.Filter(self)
                new_filter.set_data(*self.get_histogram_vals(self.available_parameters[0]), self.available_parameters[0])
                self.filters.append(new_filter)
            imgui.pop_style_color(2)
            super().render_end()

    def on_gain_focus(self):
        self.get_histogram_parameters()

    def get_histogram_vals(self, parameter):
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            particledata = datasource.get_particle_data()
            return particledata.histogram_counts[parameter], particledata.histogram_bins[parameter]
        else:
            return np.asarray([0, 0]).astype('float32'), np.asarray([0, 1]).astype('float32')

    def get_histogram_parameters(self):
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            particledata = datasource.get_particle_data()
            self.available_parameters = list(particledata.histogram_counts.keys())

    def get_particle_data_impl(self):
        if cfg.profiling:
            time_start = time.time()
        datasource = self.reconstruction_in.get_incoming_node()
        if datasource:
            pdata = datasource.get_particle_data()
            for pf in self.filters:
                pf.apply(pdata)
            if cfg.profiling:
                self.profiler_time += time.time() - time_start
            return pdata
        else:
            if cfg.profiling:
                self.profiler_time += time.time() - time_start
            return ParticleData()

    class Filter:
        def __init__(self, parent):
            self.parameter_key = ""
            self.vals = [0, 0]
            self.bins = [0, 0]
            self.min = 0
            self.max = 1
            self.parameter = 0
            self.invert = False
            self.parent = parent

        def set_data(self, vals, bins, prm_key):
            self.vals = vals
            self.bins = bins
            self.min = bins[0]
            self.max = bins[-1]
            self.parameter_key = prm_key

        def render(self):
            content_width = imgui.get_content_region_available_width()
            imgui.plot_histogram("##histogram", self.vals, graph_size = (content_width, 40))
            imgui.text("{:.2f}".format(self.bins[0]))
            imgui.same_line(position=content_width - imgui.get_font_size() * len(str(self.bins[-1])) / 2)
            imgui.text("{:.2f}".format(self.bins[-1]))
            imgui.push_item_width(content_width)
            _c, self.min = imgui.slider_float("##min", self.min, self.bins[0], self.bins[-1], "min: %1.2f")
            _c, self.max = imgui.slider_float("##max", self.max, self.bins[0], self.bins[-1], "max: %1.2f")
            imgui.pop_item_width()
            imgui.separator()
            imgui.spacing()

        def apply(self, particle_data_object):
            if cfg.profiling:
                time_start = time.time()
            particle_data_object.apply_filter(self.parameter_key, self.min, self.max, logic_not=self.invert)
            if cfg.profiling:
                self.parent.profiler_time += time.time() - time_start



import os
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
import tifffile
import glob
from scNodes.core import config as cfg


def printProgressBar(iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    # from: https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters?noredirect=1&lq=1
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end=printEnd)


def save_tiff(array, path, pixel_size_nm = 100, axes = "ZXY"):
    if not (path[-5:] == ".tiff" or path[-4:] == ".tif"):
        path += ".tif"
    if "/" in path:
        root = path.rfind("/")
        root = path[:root]
        if not os.path.exists(root):
            os.makedirs(root)
    tifffile.imwrite(path, array.astype(np.float32), resolution=(1./(1e-7 * pixel_size_nm), 1./(1e-7 * pixel_size_nm), 'CENTIMETER'))  # Changed to astype npfloat32 on 230105 to fix importing tifffile tiff with PIL Image open. Default for tifffile export unspecified-float np array appears to be 64bit which PIL doesnt support.


def save_png(array, path, alpha=True):
    try:
        if not path[-4:] == '.png':
            path += '.png'
        if "/" in path:
            root = path.rfind("/")
            root = path[:root]
            if not os.path.exists(root):
                os.makedirs(root)
        if array.dtype != np.dtype(np.uint8):
            array = array.astype(np.uint8)
        if alpha:
            Image.fromarray(array, mode="RGBA").save(path)
        else:
            Image.fromarray(array, mode="RGB").save(path)
    except Exception as e:
        cfg.set_error(e, "Error exporting image as .png. Is the path valid?")


def plot_histogram(data, bins = 'auto', title = None):
    plt.hist(data, bins = bins)
    if title:
        plt.title(title)
    plt.show()


def get_filetype(path):
    return path[path.rfind("."):]


def apply_lut_to_float_image(image, lut, contrast_lims = None):
    if len(image.shape) != 2:
        print("Image input in apply_lut_to_float_image is not 2D.")
        return False
    if isinstance(lut, list):
        _lut = np.asarray(lut)
    L, n = np.shape(_lut)
    if contrast_lims is None:
        contrast_lims = (np.amin(image), np.amax(image))
    image = (L-1) * (image - contrast_lims[0]) / (contrast_lims[1] - contrast_lims[0])
    image = image.astype(int)
    w, h = image.shape
    out_img = np.zeros((w, h, n))
    for x in range(w):
        out_img[x, :, :] = _lut[image[x, :], :]
    return out_img

def is_path_tiff(path):
    if path[-4:] == ".tif" or path[-5:] == ".tiff":
        return True
    return False

def load(path, tag=None):
    if is_path_tiff(path):
        return loadtiff(path)
    else:
        return loadfolder(path, tag)


def loadfolder(path, tag=None):
    if path[-1] != "/":
        path += "/"
    pre_paths = glob.glob(path + "*.tiff") + glob.glob(path + "*.tif")
    paths = list()
    if tag is not None:
        for path in pre_paths:
            if tag in path:
                paths.append(path)
    else:
        paths = pre_paths
    _data = Image.open(paths[0])
    _frame = np.asarray(_data)

    width = _frame.shape[0]
    height = _frame.shape[1]
    depth = len(paths)
    data = np.zeros((depth, width, height), dtype=np.float32)
    _n = len(paths)
    for f in range(0, _n):
        printProgressBar(f, _n, prefix="Loading tiff file\t", length=30, printEnd="\r")
        _data = Image.open(paths[f])
        data[f, :, :] = np.asarray(_data)
    return data


def loadtiff(path):
    "Loads stack as 3d array with dimensions (frames, width, height)"
    _data = Image.open(path)
    _frame = np.asarray(_data)

    width = _frame.shape[0]
    height = _frame.shape[1]
    depth = _data.n_frames
    data = np.zeros((depth, width, height), dtype=np.int16)
    for f in range(0, depth):
        printProgressBar(f, depth - 1, prefix="Loading tiff file\t", length=30, printEnd="\r")
        _data.seek(f)
        data[f, :, :] = np.asarray(_data)
    data = np.squeeze(data)
    return data



from scNodes.nodes.pysofi import deconvsk, ldrc
from scNodes.nodes.pysofi import finterp as f, masks, reconstruction
from scNodes.nodes.pysofi import switches as s

if s.SPHINX_SWITCH is False:
    import tifffile as tiff

import numpy as np


class PysofiData:
    """
    Data object contains the information of a dataset (e.g. dimensions,
    frame numbers), and provides SOFI methods to perform analysis and 
    visualization on the dataset (moments reconstruction, cumulants 
    reconstruction, shrinking kernel deconvolution, etc...). 

    When loading a tiff file, a PysofiData() object is created and further 
    SOFI analysis can be preformed. All the standard data-attributes 
    are listed below. New data-attributes can be added or updated using 
    '.add()' function.

    Parameters
    ----------
    filapath : str
        Path to the tiff file.
    filename : str
        Name of the tiff file.

    Attributes
    ----------
    filename : str
    filepath : str
    ave : ndarray
        The average image of the image stack / tiff video.
    finterp_factor : int
        The interpolation factor for Fourier interpolation.
    morder_lst : list
        All orders of moments-reconstructions that have been calculated.
    morder_finterp_lst : list
        All orders of moments-reconstructions after Fourier interpolation
        that have been calculated.
    moments_set : dict
        moment order (int) -> moment-reconstructed image (ndarray)
        A dictionary of orders and corrensponding moment reconstructions.
    moments_set_bc : dict
        moment order (int) -> moment-reconstructed image (ndarray)
        A dictionary of orders and corrensponding moment reconstructions 
        with bleaching correction.
    cumulants_set : dict
        cumulant order (int) -> cumulant-reconstructed image (ndarray)
        A dictionary of orders and corrensponding cumulant reconstructions.  
    cumulants_set_bc : dict 
        cumulant order (int) -> cumulant-reconstructed image (ndarray)
        A dictionary of orders and corrensponding cumulant reconstructions 
        with bleaching correction.   
    morder : int
        The highest order of moment reconstruction that has been calculated.
    corder : int
        The highest order of cumulant reconstruction that has been 
        calculated.  
    fbc : float
        Bleaching correction factor.
    n_frames : int
        The number of frames of the image stack / tiff video.
    xdim : int
        The number of pixels in x dimension.
    ydim : int
        The number of pixels in y dimension.


    Notes
    -----
    For SOFI processing, after loading the tiff video into the Data object, 
    a pipeline of 1) fourier interpolation, 2) moments reconstrtuction or 
    cumulants reconstruction, 3) noise filtering 1, 4) shrinking kernel de-
    convolution, 5) noise filtering 2, and 6) ldrc. The processed video will
    be saved into a new tiff file with the colormap user selects.

    References
    ----------
    .. [1] Xiyu Yi, Sungho Son, Ryoko Ando, Atsushi Miyawaki, and Shimon Weiss, 
    "Moments reconstruction and local dynamic range compression of high order 
    superresolution optical fluctuation imaging," Biomed. Opt. Express 10, 
    2430-2445 (2019).
    """

    def __init__(self, filepath, filename):
        self.filename = filename
        self.filepath = filepath
        self.ave = None
        self.finterp_factor = 1
        self.morder_lst = []
        self.morder_finterp_lst = []
        self.moments_set = {}
        self.moments_set_bc = {}
        self.moments_finterp_set = {}
        self.cumulants_set = {}
        self.cumulants_set_bc = {}
        self.morder = 0
        self.corder = 0
        self.fbc = 1
        self.n_frames, self.xdim, self.ydim = self.get_dims()

    def average_image(self, frames=[]):
        """Calculate average image of the tiff video."""
        self.ave = reconstruction.average_image(self.filepath, self.filename, frames)
        return self.ave

    def average_image_with_finterp(self, interp_num):
        if self.ave is not None:
            finterp_ave = f.fourier_interp_array(self.ave, [interp_num])
            return finterp_ave[0]
        else:
            finterp_ave = reconstruction.average_image_with_finterp(
                self.filepath, self.filename, interp_num)
            return finterp_ave

    def moment_image(self, order=6, mean_im=None, mvlength=[],
                     finterp=False, interp_num=1):
        """
        Calculate the moment-reconstructed image of a defined order.

        Parameters
        ----------
        order : int
            The order number of the moment-reconstructed image.
        mean_im : ndarray
            Average image of the tiff stack.
        mvlength : int
            Length of the video to calculate moments-reconstruction.
        finterp : bool
            Whether to first conduct Fourier interpolation then calculate
            moments-reconstructions.
        interp_num : int
            The interpolation factor.

        Returns
        -------
        moment_im : ndarray
            The calculated moment-reconstructed image.
        """
        if finterp is False:
            if order in self.morder_lst:
                print("this order has been calculated")
                print('\n')
            moment_im = reconstruction.calc_moment_im(self.filepath,
                                                      self.filename,
                                                      order, mvlength)
            self.morder_lst.append(order)
            self.moments_set[order] = moment_im
            return self.moments_set[order]
        else:
            if self.finterp_factor != 1 and interp_num != self.finterp_factor:
                print('Different interpolation factor calculating ...')
                print('\n')
            else:
                if order in self.morder_finterp_lst:
                    print("this order has been calculated")
                    print('\n')
            moment_im = reconstruction.moment_im_with_finterp(self.filepath,
                                                              self.filename,
                                                              order, interp_num,
                                                              mvlength)
            # print(np.shape(moment_im))
            self.morder_finterp_lst.append(order)
            self.moments_finterp_set[order] = moment_im
            self.finterp_factor = interp_num
            return self.moments_finterp_set[order]

    def calc_moments_set(self, highest_order=4, frames=[],
                         mean_im=None, bleach_correction=False,
                         smooth_kernel=251, fbc=0.04):
        """
        Calculate moment-reconstructed images to the highest order.

        Parameters
        ----------
        highest_order : int
            The highest order number of moment-reconstructed images.
        frames : list of int
            The start and end frame number.
        mean_im : ndarray
            Average image of the tiff stack.
        bleach_correction : bool
            Whether to use bleaching correction.
        smooth_kernel : int, optional
            The size of the median filter window. Only used when bleach
            correction is True.
        fbc : float, optional
            The fraction of signal decrease within each block compared
            to the total signal decrease. Only used when bleach correction
            is True.

        Returns
        -------
        moments_set, moments_set_bc : dict
            order number (int) -> image (ndarray)
            A dictionary of calcualted moment-reconstructed images.
        """
        if bleach_correction is False:
            self.moments_set = reconstruction.calc_moments(self.filepath,
                                                           self.filename,
                                                           highest_order,
                                                           frames,
                                                           m_set={})
            self.morder = highest_order
            return self.moments_set
        else:
            self.moments_set_bc = reconstruction.block_ave_moments(
                self.filepath, self.filename,
                highest_order, smooth_kernel, fbc)
            self.fbc = fbc
            return self.moments_set_bc

    def cumulants_images(self, highest_order=4, frames=[],
                         m_set=None, bleach_correction=False,
                         smooth_kernel=251, fbc=0.04):
        """
        Calculate cumulant-reconstructed images to the highest order.

        Parameters
        ----------
        highest_order : int
            The highest order number of cumulant-reconstructed images.
        frames : list of int
            The start and end frame number.
        m_set : dict
            order number (int) -> image (ndarray)
            A dictionary of calcualted moment-reconstructed images.
        bleach_correction : bool
            Whether to use bleaching correction.
        smooth_kernel : int, optional
            The size of the median filter window. Only used when bleach
            correction is True.
        fbc : float, optional
            The fraction of signal decrease within each block compared
            to the total signal decrease. Only used when bleach correction
            is True.

        Returns
        -------
        cumulants_set, cumulants_set_bc : dict
            order number (int) -> image (ndarray)
            A dictionary of calcualted cumulant-reconstructed images.
        """
        if bleach_correction is False:
            self.corder = highest_order
            m_set = self.calc_moments_set(highest_order, frames)
            self.cumulants_set = reconstruction.calc_cumulants_from_moments(
                m_set)
            return self.cumulants_set
        else:
            self.cumulants_set_bc = reconstruction.block_ave_cumulants(
                self.filepath, self.filename,
                highest_order, smooth_kernel, fbc)
            return self.cumulants_set_bc

    def ldrc(self, order=6, window_size=[25, 25], mask_im=None, input_im=None):
        """
        Compress the dynamic range of moments-/cumulants reconstruced image
        with ldrc method. For details of ldrc, refer to ldrc.py and [1].

        Parameters
        ----------
        order : int
            The order of the reconstructed image.
        window_size : [int, int]
            The [x, y] dimension of the scanning window.
        mask_im : ndarray
            A reference image.
            Usually a average/sum image or second-order SOFI image is used.
        input_im : ndarray
            An input image, usually a high-order moment- or cumulant-
            reconstructed image.

        Returns
        -------
        ldrc_image : ndarray
            The compressed image with the same dimensions of input_im.
        """
        if input_im is None:
            if self.moments_set is not None:
                if self.morder >= order:
                    input_im = self.moments_set[order]
                else:
                    self.moment_image(order)
            else:
                self.moment_image(order)

        if mask_im is None:
            mask_im = self.average_image()

        self.ldrc_image = ldrc.ldrc(mask_im, input_im, order, window_size)
        return self.ldrc_image

    def deconvsk(self, est_psf=masks.gauss2d_mask((51, 51), 2), input_im=None,
                 deconv_lambda=1.5, deconv_iter=20):
        """
        Perform serial Richardson-Lucy deconvolution with a series of PSFs
        on the input image. For details of shrinking kernel deconvolution,
        refer to deconvsk.py and [1].

        Parameters
        ----------
        est_psf : ndarray
            Estimated PSF.
        input_im : ndarray
            Input image that need deconvolution.
        deconv_lambda : float
            Lambda for the exponent between. It is an empirical parameter
            within the range of (1,2).
        deconv_iter : int
            Number of iterations for each deconvolution.

        Returns
        -------
        deconv : ndarray
            Deconvoluted image.
        """
        if input_im is None:
            if self.moments_set is not None:
                input_im = self.moments_set[self.morder]
            elif self.ave is not None:
                input_im = self.ave
            else:
                input_im = self.average_image()
        self.deconv = deconvsk.deconvsk(est_psf, input_im,
                                        deconv_lambda, deconv_iter)
        return self.deconv

    def finterp_tiffstack(self, interp_num_lst=[2, 4], frames=[],
                          save_option=True, return_option=False):
        """
        Performs fourier interpolation on a tiff image (stack) with a list of
        interpolation factors (the number of pixels to be interpolated between
        two adjacent pixels).

        Parameters
        ----------
        interp_num_lst : list (int)
            A list of interpolation factors.
        frames : list of int
            The start and end frame number.
        save_option : bool
            Whether to save the interpolated images into tiff files (each
            interpolation factor seperately).
        return_option : bool
            Whether to return the interpolated image series as 3d arrays.

        Returns
        -------
        finterps : list (ndarray)
            A list of interpolated image sereis corresponding to the interpolation
            factor list.
        """
        filename = self.filename[:-4]
        if return_option is False:
            f.fourier_interp_tiff(self.filepath, filename,
                                  interp_num_lst, frames,
                                  save_option, return_option)
        else:
            self.finterps = f.fourier_interp_tiff(self.filepath,
                                                  filename, interp_num_lst,
                                                  frames, save_option,
                                                  return_option)
            return self.finterps

    def finterp_image(self, input_im=None, interp_num_lst=[2, 4]):
        """Performs fourier interpolation on an image array."""
        if input_im is None:
            input_im = self.average_image()

        self.finterp = f.fourier_interp_array(input_im, interp_num_lst)
        return self.finterp

    def bleach_correct(self, fbc=0.04, smooth_kernel=251,
                       save_option=True, return_option=False):
        """
        Performs bleaching correction on a tiff image (stack) with a
        bleaching correction factor.

        Parameters
        ----------
        fbc : float
            The fraction of signal decrease within each block compared
            to the total signal decrease. Only used when bleach correction
            is True.
        smooth_kernel : int
            The size of the median filter window.
        save_option : bool
            Whether to save the corrected images into tiff files.
        return_option : bool
            Whether to return the corrected image series as a 3d array.

        Returns
        -------
        bc_im : ndarray
            All bleaching-corrected imagee in a 3d array.
        """
        if save_option is True:
            reconstruction.correct_bleaching(self.filepath, self.filename,
                                             fbc, smooth_kernel,
                                             save_option, return_option)
        if return_option is True:
            self.bc_ims = reconstruction.correct_bleaching(self.filepath,
                                                           self.filename,
                                                           fbc,
                                                           smooth_kernel,
                                                           save_option,
                                                           return_option)
            return self.bc_ims

    def add_filtered(self, image, filter_name='noise filter'):
        """Add (noise) filtered image as an attribute of the object."""
        self.filter_name = filter_name
        self.filtered = image

    def get_dims(self):
        """Get dimensions and frame number of the tiff video."""
        imstack = tiff.TiffFile(self.filepath + '/' + self.filename)
        n_frames = len(imstack.pages)
        xdim, ydim = np.shape(imstack.pages[0])
        return n_frames, xdim, ydim

    def add(self, **kwargs):
        """Add or update elements (attributes and/or dict entries). """
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_frame(self, frame_num=0):
        """
        Get one frame of the tiff video.

        Parameters
        ----------
        frame_num : int
            The frame number (e.g. 0 for the first frame).

        Returns
        -------
        frame_im : ndarray
            Frame array.
        """
        if frame_num >= self.n_frames:
            raise Exception("'frame_num' exceeds the length of the video")
        if frame_num < 0 or np.int(frame_num) != frame_num:
            raise Exception("'frame_num' should be a non-negative integer")
        frame_im = tiff.imread(self.filepath + '/' +
                               self.filename, key=frame_num)
        return frame_im